import json
import unittest
from base64 import b64encode
from datetime import datetime, timezone
from unittest.mock import ANY

from django.test import tag

from anymail.exceptions import AnymailConfigurationError
from anymail.signals import AnymailTrackingEvent
from anymail.webhooks.postal import PostalTrackingWebhookView

from .utils_postal import ClientWithPostalSignature, make_key
from .webhook_cases import WebhookTestCase


@tag("postal")
@unittest.skipUnless(
    ClientWithPostalSignature, "Install 'cryptography' to run postal webhook tests"
)
class PostalWebhookSecurityTestCase(WebhookTestCase):
    client_class = ClientWithPostalSignature

    def setUp(self):
        super().setUp()
        self.clear_basic_auth()

        self.client.set_private_key(make_key())

    def test_failed_signature_check(self):
        response = self.client.post(
            "/anymail/postal/tracking/",
            content_type="application/json",
            data=json.dumps({"some": "data"}),
            HTTP_X_POSTAL_SIGNATURE=b64encode("invalid".encode("utf-8")),
        )
        self.assertEqual(response.status_code, 400)

        response = self.client.post(
            "/anymail/postal/tracking/",
            content_type="application/json",
            data=json.dumps({"some": "data"}),
            HTTP_X_POSTAL_SIGNATURE="garbage",
        )
        self.assertEqual(response.status_code, 400)

        response = self.client.post(
            "/anymail/postal/tracking/",
            content_type="application/json",
            data=json.dumps({"some": "data"}),
            HTTP_X_POSTAL_SIGNATURE="",
        )
        self.assertEqual(response.status_code, 400)


@tag("postal")
@unittest.skipUnless(
    ClientWithPostalSignature, "Install 'cryptography' to run postal webhook tests"
)
class PostalDeliveryTestCase(WebhookTestCase):
    client_class = ClientWithPostalSignature

    def setUp(self):
        super().setUp()
        self.clear_basic_auth()

        self.client.set_private_key(make_key())

    def test_bounce_event(self):
        raw_event = {
            "event": "MessageDelayed",
            "timestamp": 1606753101.961181,
            "payload": {
                "original_message": {
                    "id": 233843,
                    "token": "McC2tuqg7mhx",
                    "direction": "outgoing",
                    "message_id": "7b82aac4-5d63-41b8-8e35-9faa31a892dc"
                    "@rp.postal.example.com",
                    "to": "bounce@example.com",
                    "from": "sender@example.com",
                    "subject": "...",
                    "timestamp": 1606436187.8883688,
                    "spam_status": "NotChecked",
                    "tag": None,
                },
                "bounce": {
                    "id": 233864,
                    "token": "nII5p0Cp8onV",
                    "direction": "incoming",
                    "message_id": "E1kiRR8-0001ay-Iq@example.com",
                    "to": "bk87jw@psrp.postal.example.com",
                    "from": None,
                    "subject": "Mail delivery failed: returning message to sender",
                    "timestamp": 1606436523.6060522,
                    "spam_status": "NotChecked",
                    "tag": None,
                },
                "details": "details",
                "output": "server output",
                "sent_with_ssl": None,
                "timestamp": 1606753101.9110143,
                "time": None,
            },
            "uuid": "0fcc831f-92b9-4e2b-97f2-d873abc77fab",
        }

        response = self.client.post(
            "/anymail/postal/tracking/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=PostalTrackingWebhookView,
            event=ANY,
            esp_name="Postal",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "bounced")
        self.assertEqual(event.esp_event, raw_event)
        self.assertEqual(
            event.timestamp, datetime.fromtimestamp(1606753101, tz=timezone.utc)
        )
        self.assertEqual(event.message_id, 233843)
        self.assertEqual(event.event_id, "0fcc831f-92b9-4e2b-97f2-d873abc77fab")
        self.assertEqual(event.recipient, "bounce@example.com")
        self.assertEqual(event.reject_reason, "bounced")
        self.assertEqual(event.description, "details")
        self.assertEqual(event.mta_response, "server output")

    def test_deferred_event(self):
        raw_event = {
            "event": "MessageDelayed",
            "timestamp": 1606753101.961181,
            "payload": {
                "message": {
                    "id": 1564,
                    "token": "Kmo8CRdjuM7B",
                    "direction": "outgoing",
                    "message_id": "7b095c0e-2c98-4e68-a41f-7bd217a83925"
                    "@rp.postal.example.com",
                    "to": "deferred@example.com",
                    "from": "test@postal.example.com",
                    "subject": "Test Message at November 30, 2020 16:03",
                    "timestamp": 1606752235.195664,
                    "spam_status": "NotChecked",
                    "tag": None,
                },
                "status": "SoftFail",
                "details": "details",
                "output": "server output",
                "sent_with_ssl": None,
                "timestamp": 1606753101.9110143,
                "time": None,
            },
            "uuid": "0fcc831f-92b9-4e2b-97f2-d873abc77fab",
        }
        response = self.client.post(
            "/anymail/postal/tracking/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=PostalTrackingWebhookView,
            event=ANY,
            esp_name="Postal",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "deferred")
        self.assertEqual(event.esp_event, raw_event)
        self.assertEqual(
            event.timestamp, datetime.fromtimestamp(1606753101, tz=timezone.utc)
        )
        self.assertEqual(event.message_id, 1564)
        self.assertEqual(event.event_id, "0fcc831f-92b9-4e2b-97f2-d873abc77fab")
        self.assertEqual(event.recipient, "deferred@example.com")
        self.assertEqual(event.reject_reason, None)
        self.assertEqual(event.description, "details")
        self.assertEqual(event.mta_response, "server output")

    def test_queued_event(self):
        raw_event = {
            "event": "MessageHeld",
            "timestamp": 1606753101.330977,
            "payload": {
                "message": {
                    "id": 1568,
                    "token": "VRvQMS20Bb4Y",
                    "direction": "outgoing",
                    "message_id": "ec7b6375-4045-451a-9503-2a23a607c1c1"
                    "@rp.postal.example.com",
                    "to": "suppressed@example.com",
                    "from": "test@example.com",
                    "subject": "Test Message at November 30, 2020 16:12",
                    "timestamp": 1606752750.993815,
                    "spam_status": "NotChecked",
                    "tag": None,
                },
                "status": "Held",
                "details": "Recipient (suppressed@example.com)"
                " is on the suppression list",
                "output": "server output",
                "sent_with_ssl": None,
                "timestamp": 1606752751.8933666,
                "time": None,
            },
            "uuid": "9be13015-2e54-456c-bf66-eacbe33da824",
        }
        response = self.client.post(
            "/anymail/postal/tracking/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=PostalTrackingWebhookView,
            event=ANY,
            esp_name="Postal",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "queued")
        self.assertEqual(event.esp_event, raw_event)
        self.assertEqual(
            event.timestamp, datetime.fromtimestamp(1606753101, tz=timezone.utc)
        )
        self.assertEqual(event.message_id, 1568)
        self.assertEqual(event.event_id, "9be13015-2e54-456c-bf66-eacbe33da824")
        self.assertEqual(event.recipient, "suppressed@example.com")
        self.assertEqual(event.reject_reason, None)
        self.assertEqual(
            event.description,
            "Recipient (suppressed@example.com) is on the suppression list",
        )
        self.assertEqual(event.mta_response, "server output")

    def test_failed_event(self):
        raw_event = {
            "event": "MessageDeliveryFailed",
            "timestamp": 1606753101.084981,
            "payload": {
                "message": {
                    "id": 1571,
                    "token": "MzWWQPubXXWz",
                    "direction": "outgoing",
                    "message_id": "cfb29da8ed1e4ed5a6c8a0f24d7a9ef3"
                    "@rp.postal.example.com",
                    "to": "failed@example.com",
                    "from": "test@example.com",
                    "subject": "Message delivery failed...",
                    "timestamp": 1606753318.072171,
                    "spam_status": "NotChecked",
                    "tag": None,
                },
                "status": "HardFail",
                "details": "Could not deliver",
                "output": "server output",
                "sent_with_ssl": None,
                "timestamp": 1606753318.7010343,
                "time": None,
            },
            "uuid": "5fec5077-dae7-4989-94d5-e1963f3e9181",
        }
        response = self.client.post(
            "/anymail/postal/tracking/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=PostalTrackingWebhookView,
            event=ANY,
            esp_name="Postal",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "failed")
        self.assertEqual(event.esp_event, raw_event)
        self.assertEqual(
            event.timestamp, datetime.fromtimestamp(1606753101, tz=timezone.utc)
        )
        self.assertEqual(event.message_id, 1571)
        self.assertEqual(event.event_id, "5fec5077-dae7-4989-94d5-e1963f3e9181")
        self.assertEqual(event.recipient, "failed@example.com")
        self.assertEqual(event.reject_reason, None)
        self.assertEqual(event.description, "Could not deliver")
        self.assertEqual(event.mta_response, "server output")

    def test_delivered_event(self):
        raw_event = {
            "event": "MessageSent",
            "timestamp": 1606753101.354368,
            "payload": {
                "message": {
                    "id": 1563,
                    "token": "zw6psSlgo6ki",
                    "direction": "outgoing",
                    "message_id": "c462ad36-be49-469c-b7b2-dfd317eb40fa"
                    "@rp.postal.example.com",
                    "to": "recipient@example.com",
                    "from": "test@example.com",
                    "subject": "Test Message at November 30, 2020 16:01",
                    "timestamp": 1606752104.699201,
                    "spam_status": "NotChecked",
                    "tag": "welcome-email",
                },
                "status": "Sent",
                "details": "Message for recipient@example.com accepted",
                "output": "250 2.0.0 OK\n",
                "sent_with_ssl": False,
                "timestamp": 1606752106.9858062,
                "time": 0.89,
            },
            "uuid": "58e8d7ee-2cd5-4db2-9af3-3f436105795a",
        }
        response = self.client.post(
            "/anymail/postal/tracking/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=PostalTrackingWebhookView,
            event=ANY,
            esp_name="Postal",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "delivered")
        self.assertEqual(event.esp_event, raw_event)
        self.assertEqual(
            event.timestamp, datetime.fromtimestamp(1606753101, tz=timezone.utc)
        )
        self.assertEqual(event.message_id, 1563)
        self.assertEqual(event.recipient, "recipient@example.com")
        self.assertEqual(event.tags, ["welcome-email"])
        self.assertEqual(event.metadata, None)

    def test_ignore_incoming_events(self):
        raw_event = {
            "event": "MessageDeliveryFailed",
            "timestamp": 1606756014.694645,
            "payload": {
                "message": {
                    "id": 1575,
                    "token": "lPDuNhHfV8aU",
                    "direction": "incoming",
                    "message_id": "asdf@other-mta.example.com",
                    "to": "incoming@example.com",
                    "from": "sender@example.com",
                    "subject": "test",
                    "timestamp": 1606756008.718169,
                    "spam_status": "NotSpam",
                    "tag": None,
                },
                "status": "HardFail",
                "details": "Received a 400 from https://anymail.example.com/"
                "anymail/postal/tracking/.",
                "output": "Not found",
                "sent_with_ssl": False,
                "timestamp": 1606756014.1078613,
                "time": 0.15,
            },
            "uuid": "a01724c0-0d1a-4090-89aa-c3da5a683375",
        }
        response = self.client.post(
            "/anymail/postal/tracking/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(self.tracking_handler.call_count, 0)

    def test_misconfigured_inbound(self):
        with self.assertRaisesMessage(
            AnymailConfigurationError,
            "You seem to have set Postal's *inbound* webhook"
            " to Anymail's Postal *tracking* webhook URL.",
        ):
            self.client.post(
                "/anymail/postal/tracking/",
                content_type="application/json",
                data=json.dumps({"rcpt_to": "to@example.org"}),
            )



import json
from base64 import b64encode
from datetime import date, datetime
from decimal import Decimal
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage

from django.core import mail
from django.core.exceptions import ImproperlyConfigured
from django.test import SimpleTestCase, override_settings, tag
from django.utils.timezone import (
    get_fixed_timezone,
    override as override_current_timezone,
)

from anymail.exceptions import (
    AnymailAPIError,
    AnymailSerializationError,
    AnymailUnsupportedFeature,
)
from anymail.message import AnymailMessage, attach_inline_image_file

from .mock_requests_backend import (
    RequestsBackendMockAPITestCase,
    SessionSharingTestCases,
)
from .utils import (
    SAMPLE_IMAGE_FILENAME,
    AnymailTestMixin,
    decode_att,
    sample_image_content,
    sample_image_path,
)


@tag("resend")
@override_settings(
    EMAIL_BACKEND="anymail.backends.resend.EmailBackend",
    ANYMAIL={
        "RESEND_API_KEY": "test_api_key",
    },
)
class ResendBackendMockAPITestCase(RequestsBackendMockAPITestCase):
    DEFAULT_RAW_RESPONSE = b'{"id": "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee"}'

    def setUp(self):
        super().setUp()
        # Simple message useful for many tests
        self.message = mail.EmailMultiAlternatives(
            "Subject", "Text Body", "from@example.com", ["to@example.com"]
        )


@tag("resend")
class ResendBackendStandardEmailTests(ResendBackendMockAPITestCase):
    """Test backend support for Django standard email features"""

    def test_send_mail(self):
        """Test basic API for simple send"""
        mail.send_mail(
            "Subject here",
            "Here is the message.",
            "from@sender.example.com",
            ["to@example.com"],
            fail_silently=False,
        )
        self.assert_esp_called("/emails")
        headers = self.get_api_call_headers()
        self.assertEqual(headers["Authorization"], "Bearer test_api_key")
        data = self.get_api_call_json()
        self.assertEqual(data["subject"], "Subject here")
        self.assertEqual(data["text"], "Here is the message.")
        self.assertEqual(data["from"], "from@sender.example.com")
        self.assertEqual(data["to"], ["to@example.com"])

    def test_name_addr(self):
        """Make sure RFC2822 name-addr format (with display-name) is allowed

        (Test both sender and recipient addresses)
        """
        msg = mail.EmailMessage(
            "Subject",
            "Message",
            "From Name <from@example.com>",
            ["Recipient #1 <to1@example.com>", "to2@example.com"],
            cc=["Carbon Copy <cc1@example.com>", "cc2@example.com"],
            bcc=["Blind Copy <bcc1@example.com>", "bcc2@example.com"],
        )
        msg.send()
        data = self.get_api_call_json()
        self.assertEqual(data["from"], "From Name <from@example.com>")
        self.assertEqual(
            data["to"], ["Recipient #1 <to1@example.com>", "to2@example.com"]
        )
        self.assertEqual(
            data["cc"], ["Carbon Copy <cc1@example.com>", "cc2@example.com"]
        )
        self.assertEqual(
            data["bcc"], ["Blind Copy <bcc1@example.com>", "bcc2@example.com"]
        )

    def test_email_message(self):
        email = mail.EmailMessage(
            "Subject",
            "Body goes here",
            "from@example.com",
            ["to1@example.com", "Also To <to2@example.com>"],
            bcc=["bcc1@example.com", "Also BCC <bcc2@example.com>"],
            cc=["cc1@example.com", "Also CC <cc2@example.com>"],
            reply_to=["another@example.com"],
            headers={
                "X-MyHeader": "my value",
            },
        )
        email.send()
        data = self.get_api_call_json()
        self.assertEqual(data["subject"], "Subject")
        self.assertEqual(data["text"], "Body goes here")
        self.assertEqual(data["from"], "from@example.com")
        self.assertEqual(data["to"], ["to1@example.com", "Also To <to2@example.com>"])
        self.assertEqual(
            data["bcc"], ["bcc1@example.com", "Also BCC <bcc2@example.com>"]
        )
        self.assertEqual(data["cc"], ["cc1@example.com", "Also CC <cc2@example.com>"])
        self.assertEqual(data["reply_to"], ["another@example.com"])
        self.assertCountEqual(
            data["headers"],
            {"X-MyHeader": "my value"},
        )

    def test_html_message(self):
        text_content = "This is an important message."
        html_content = "<p>This is an <strong>important</strong> message.</p>"
        email = mail.EmailMultiAlternatives(
            "Subject", text_content, "from@example.com", ["to@example.com"]
        )
        email.attach_alternative(html_content, "text/html")
        email.send()
        data = self.get_api_call_json()
        self.assertEqual(data["text"], text_content)
        self.assertEqual(data["html"], html_content)
        # Don't accidentally send the html part as an attachment:
        self.assertNotIn("attachments", data)

    def test_html_only_message(self):
        html_content = "<p>This is an <strong>important</strong> message.</p>"
        email = mail.EmailMessage(
            "Subject", html_content, "from@example.com", ["to@example.com"]
        )
        email.content_subtype = "html"  # Main content is now text/html
        email.send()
        data = self.get_api_call_json()
        self.assertNotIn("text", data)
        self.assertEqual(data["html"], html_content)

    def test_extra_headers(self):
        self.message.extra_headers = {"X-Custom": "string", "X-Num": 123}
        self.message.send()
        data = self.get_api_call_json()
        # header values must be strings (or they'll cause an "invalid literal" API error)
        self.assertEqual(data["headers"], {"X-Custom": "string", "X-Num": "123"})

    def test_extra_headers_serialization_error(self):
        self.message.extra_headers = {"X-Custom": Decimal(12.5)}
        with self.assertRaisesMessage(AnymailSerializationError, "Decimal"):
            self.message.send()

    def test_reply_to(self):
        email = mail.EmailMessage(
            "Subject",
            "Body goes here",
            "from@example.com",
            ["to1@example.com"],
            reply_to=["reply@example.com", "Other <reply2@example.com>"],
        )
        email.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["reply_to"], ["reply@example.com", "Other <reply2@example.com>"]
        )

    def test_attachments(self):
        text_content = "* Item one\n* Item two\n* Item three"
        self.message.attach(
            filename="test.txt", content=text_content, mimetype="text/plain"
        )

        # Should guess mimetype if not provided...
        png_content = b"PNG\xb4 pretend this is the contents of a png file"
        self.message.attach(filename="test.png", content=png_content)

        # Should work with a MIMEBase object (also tests no filename)...
        pdf_content = b"PDF\xb4 pretend this is valid pdf data"
        mimeattachment = MIMEBase("application", "pdf")
        mimeattachment.set_payload(pdf_content)
        self.message.attach(mimeattachment)

        self.message.send()
        data = self.get_api_call_json()
        attachments = data["attachments"]
        self.assertEqual(len(attachments), 3)
        self.assertEqual(attachments[0]["filename"], "test.txt")
        self.assertEqual(
            decode_att(attachments[0]["content"]).decode("ascii"), text_content
        )

        self.assertEqual(attachments[1]["filename"], "test.png")
        self.assertEqual(decode_att(attachments[1]["content"]), png_content)

        # unnamed attachment given default name with correct extension for content type
        self.assertEqual(attachments[2]["filename"], "attachment.pdf")
        self.assertEqual(decode_att(attachments[2]["content"]), pdf_content)

    def test_unicode_attachment_correctly_decoded(self):
        self.message.attach(
            "Une pièce jointe.html", "<p>\u2019</p>", mimetype="text/html"
        )
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["attachments"],
            [
                {
                    "filename": "Une pièce jointe.html",
                    "content": b64encode("<p>\u2019</p>".encode("utf-8")).decode(
                        "ascii"
                    ),
                }
            ],
        )

    def test_embedded_images(self):
        # Resend's API doesn't have a way to specify content-id
        image_filename = SAMPLE_IMAGE_FILENAME
        image_path = sample_image_path(image_filename)

        cid = attach_inline_image_file(self.message, image_path)  # Read from a png file
        html_content = (
            '<p>This has an <img src="cid:%s" alt="inline" /> image.</p>' % cid
        )
        self.message.attach_alternative(html_content, "text/html")

        with self.assertRaisesMessage(AnymailUnsupportedFeature, "inline content-id"):
            self.message.send()

    def test_attached_images(self):
        image_filename = SAMPLE_IMAGE_FILENAME
        image_path = sample_image_path(image_filename)
        image_data = sample_image_content(image_filename)

        # option 1: attach as a file
        self.message.attach_file(image_path)

        # option 2: construct the MIMEImage and attach it directly
        image = MIMEImage(image_data)
        self.message.attach(image)

        image_data_b64 = b64encode(image_data).decode("ascii")

        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["attachments"],
            [
                {
                    "filename": image_filename,  # the named one
                    "content": image_data_b64,
                },
                {
                    # For unnamed attachments, Anymail constructs a default name
                    # based on the content_type:
                    "filename": "attachment.png",
                    "content": image_data_b64,
                },
            ],
        )

    def test_multiple_html_alternatives(self):
        # Multiple alternatives not allowed
        self.message.attach_alternative("<p>First html is OK</p>", "text/html")
        self.message.attach_alternative("<p>But not second html</p>", "text/html")
        with self.assertRaisesMessage(AnymailUnsupportedFeature, "multiple html parts"):
            self.message.send()

    def test_html_alternative(self):
        # Only html alternatives allowed
        self.message.attach_alternative("{'not': 'allowed'}", "application/json")
        with self.assertRaises(AnymailUnsupportedFeature):
            self.message.send()

    def test_alternatives_fail_silently(self):
        # Make sure fail_silently is respected
        self.message.attach_alternative("{'not': 'allowed'}", "application/json")
        sent = self.message.send(fail_silently=True)
        self.assert_esp_not_called("API should not be called when send fails silently")
        self.assertEqual(sent, 0)

    def test_suppress_empty_address_lists(self):
        """Empty to, cc, bcc, and reply_to shouldn't generate empty fields"""
        self.message.send()
        data = self.get_api_call_json()
        self.assertNotIn("cc", data)
        self.assertNotIn("bcc", data)
        self.assertNotIn("reply_to", data)

        # Test empty `to`--but send requires at least one recipient somewhere (like cc)
        self.message.to = []
        self.message.cc = ["cc@example.com"]
        self.message.send()
        data = self.get_api_call_json()
        self.assertNotIn("to", data)

    def test_api_failure(self):
        failure_response = {
            "statusCode": 400,
            "message": "API key is invalid",
            "name": "validation_error",
        }
        self.set_mock_response(status_code=400, json_data=failure_response)
        with self.assertRaisesMessage(
            AnymailAPIError, r"Resend API response 400"
        ) as cm:
            mail.send_mail("Subject", "Body", "from@example.com", ["to@example.com"])
        self.assertIn("API key is invalid", str(cm.exception))

        # Make sure fail_silently is respected
        self.set_mock_response(status_code=422, json_data=failure_response)
        sent = mail.send_mail(
            "Subject",
            "Body",
            "from@example.com",
            ["to@example.com"],
            fail_silently=True,
        )
        self.assertEqual(sent, 0)


@tag("resend")
class ResendBackendAnymailFeatureTests(ResendBackendMockAPITestCase):
    """Test backend support for Anymail added features"""

    def test_envelope_sender(self):
        self.message.envelope_sender = "anything@bounces.example.com"
        with self.assertRaisesMessage(AnymailUnsupportedFeature, "envelope_sender"):
            self.message.send()

    def test_metadata(self):
        self.message.metadata = {"user_id": "12345", "items": 6}
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            json.loads(data["headers"]["X-Metadata"]),
            {"user_id": "12345", "items": 6},
        )

    def test_send_at(self):
        utc_plus_6 = get_fixed_timezone(6 * 60)
        utc_minus_8 = get_fixed_timezone(-8 * 60)

        with override_current_timezone(utc_plus_6):
            # Timezone-naive datetime assumed to be Django current_timezone
            self.message.send_at = datetime(2022, 10, 11, 12, 13, 14, 123456)
            self.message.send()
            data = self.get_api_call_json()
            # (Resend can't handle microseconds; truncate to milliseconds.)
            self.assertEqual(data["scheduled_at"], "2022-10-11T12:13:14.123+06:00")

            # Timezone-aware datetime converted to UTC:
            self.message.send_at = datetime(2016, 3, 4, 5, 6, 7, tzinfo=utc_minus_8)
            self.message.send()
            data = self.get_api_call_json()
            self.assertEqual(data["scheduled_at"], "2016-03-04T05:06:07-08:00")

            # Date-only treated as midnight in current timezone
            # (which probably won't send since it's not in the future)
            self.message.send_at = date(2022, 10, 22)
            self.message.send()
            data = self.get_api_call_json()
            self.assertEqual(data["scheduled_at"], "2022-10-22T00:00:00+06:00")

            # POSIX timestamp
            self.message.send_at = 1651820889  # 2022-05-06 07:08:09 UTC
            self.message.send()
            data = self.get_api_call_json()
            self.assertEqual(data["scheduled_at"], "2022-05-06T07:08:09+00:00")

            # String passed unchanged (this is *not* portable between ESPs)
            self.message.send_at = "2013-11-12T01:02:03Z"
            self.message.send()
            data = self.get_api_call_json()
            self.assertEqual(data["scheduled_at"], "2013-11-12T01:02:03Z")

    def test_tags(self):
        self.message.tags = ["receipt", "reorder test 12"]
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            json.loads(data["headers"]["X-Tags"]),
            ["receipt", "reorder test 12"],
        )

    def test_headers_metadata_tags_interaction(self):
        # Test three features that use custom headers don't clobber each other
        self.message.extra_headers = {"X-Custom": "custom value"}
        self.message.metadata = {"user_id": "12345"}
        self.message.tags = ["receipt", "reorder test 12"]
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["headers"],
            {
                "X-Custom": "custom value",
                "X-Tags": '["receipt", "reorder test 12"]',
                "X-Metadata": '{"user_id": "12345"}',
            },
        )

    _mock_batch_response = {
        "data": [
            {"id": "ae2014de-c168-4c61-8267-70d2662a1ce1"},
            {"id": "faccb7a5-8a28-4e9a-ac64-8da1cc3bc1cb"},
        ]
    }

    def test_merge_data(self):
        self.message.merge_data = {"to@example.com": {"customer_id": 3}}
        with self.assertRaisesMessage(AnymailUnsupportedFeature, "merge_data"):
            self.message.send()

    def test_empty_merge_data(self):
        # `merge_data = {}` triggers batch send
        self.set_mock_response(json_data=self._mock_batch_response)
        message = AnymailMessage(
            from_email="from@example.com",
            to=["alice@example.com", "Bob <bob@example.com>"],
            cc=["cc@example.com"],
            merge_data={
                "alice@example.com": {},
                "bob@example.com": {},
            },
        )
        message.send()
        self.assert_esp_called("/emails/batch")
        data = self.get_api_call_json()
        self.assertEqual(len(data), 2)
        self.assertEqual(data[0]["to"], ["alice@example.com"])
        self.assertEqual(data[1]["to"], ["Bob <bob@example.com>"])

        recipients = message.anymail_status.recipients
        self.assertEqual(recipients["alice@example.com"].status, "queued")
        self.assertEqual(
            recipients["alice@example.com"].message_id,
            "ae2014de-c168-4c61-8267-70d2662a1ce1",
        )
        self.assertEqual(recipients["bob@example.com"].status, "queued")
        self.assertEqual(
            recipients["bob@example.com"].message_id,
            "faccb7a5-8a28-4e9a-ac64-8da1cc3bc1cb",
        )
        # No message_id for cc/bcc recipients in a batch send
        self.assertEqual(recipients["cc@example.com"].status, "queued")
        self.assertIsNone(recipients["cc@example.com"].message_id)

    def test_merge_metadata(self):
        self.set_mock_response(json_data=self._mock_batch_response)
        message = AnymailMessage(
            from_email="from@example.com",
            to=["alice@example.com", "Bob <bob@example.com>"],
            merge_metadata={
                "alice@example.com": {"order_id": 123, "tier": "premium"},
                "bob@example.com": {"order_id": 678},
            },
            metadata={"notification_batch": "zx912"},
        )
        message.send()

        # merge_metadata forces batch send API:
        self.assert_esp_called("/emails/batch")

        data = self.get_api_call_json()
        self.assertEqual(len(data), 2)
        self.assertEqual(data[0]["to"], ["alice@example.com"])
        # metadata and merge_metadata[recipient] are combined:
        self.assertEqual(
            json.loads(data[0]["headers"]["X-Metadata"]),
            {"order_id": 123, "tier": "premium", "notification_batch": "zx912"},
        )
        self.assertEqual(data[1]["to"], ["Bob <bob@example.com>"])
        self.assertEqual(
            json.loads(data[1]["headers"]["X-Metadata"]),
            {"order_id": 678, "notification_batch": "zx912"},
        )

        recipients = message.anymail_status.recipients
        self.assertEqual(recipients["alice@example.com"].status, "queued")
        self.assertEqual(
            recipients["alice@example.com"].message_id,
            "ae2014de-c168-4c61-8267-70d2662a1ce1",
        )
        self.assertEqual(recipients["bob@example.com"].status, "queued")
        self.assertEqual(
            recipients["bob@example.com"].message_id,
            "faccb7a5-8a28-4e9a-ac64-8da1cc3bc1cb",
        )

    def test_merge_headers(self):
        self.set_mock_response(json_data=self._mock_batch_response)
        message = AnymailMessage(
            from_email="from@example.com",
            to=["alice@example.com", "Bob <bob@example.com>"],
            headers={
                "List-Unsubscribe-Post": "List-Unsubscribe=One-Click",
                "List-Unsubscribe": "<mailto:unsubscribe@example.com>",
            },
            merge_headers={
                "alice@example.com": {
                    "List-Unsubscribe": "<https://example.com/a/>",
                },
                "bob@example.com": {
                    "List-Unsubscribe": "<https://example.com/b/>",
                },
            },
        )
        message.send()

        # merge_headers forces batch send API:
        self.assert_esp_called("/emails/batch")

        data = self.get_api_call_json()
        self.assertEqual(len(data), 2)
        self.assertEqual(data[0]["to"], ["alice@example.com"])
        # global and recipient headers are combined:
        self.assertEqual(
            data[0]["headers"],
            {
                "List-Unsubscribe": "<https://example.com/a/>",
                "List-Unsubscribe-Post": "List-Unsubscribe=One-Click",
            },
        )
        self.assertEqual(data[1]["to"], ["Bob <bob@example.com>"])
        self.assertEqual(
            data[1]["headers"],
            {
                "List-Unsubscribe": "<https://example.com/b/>",
                "List-Unsubscribe-Post": "List-Unsubscribe=One-Click",
            },
        )

    def test_track_opens(self):
        self.message.track_opens = True
        with self.assertRaisesMessage(AnymailUnsupportedFeature, "track_opens"):
            self.message.send()

    def test_track_clicks(self):
        self.message.track_clicks = True
        with self.assertRaisesMessage(AnymailUnsupportedFeature, "track_clicks"):
            self.message.send()

    def test_default_omits_options(self):
        """Make sure by default we don't send any ESP-specific options.

        Options not specified by the caller should be omitted entirely from
        the API call (*not* sent as False or empty). This ensures
        that your ESP account settings apply by default.
        """
        self.message.send()
        data = self.get_api_call_json()
        self.assertNotIn("headers", data)
        self.assertNotIn("attachments", data)
        self.assertNotIn("tags", data)

    def test_esp_extra(self):
        self.message.esp_extra = {
            "tags": [{"name": "my_tag", "value": "my_tag_value"}],
        }
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["tags"], [{"name": "my_tag", "value": "my_tag_value"}])

    # noinspection PyUnresolvedReferences
    def test_send_attaches_anymail_status(self):
        """The anymail_status should be attached to the message when it is sent"""
        msg = mail.EmailMessage(
            "Subject",
            "Message",
            "from@example.com",
            ["Recipient <to1@example.com>"],
        )
        sent = msg.send()
        self.assertEqual(sent, 1)
        self.assertEqual(msg.anymail_status.status, {"queued"})
        self.assertEqual(
            msg.anymail_status.message_id, "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee"
        )
        self.assertEqual(
            msg.anymail_status.recipients["to1@example.com"].status, "queued"
        )
        self.assertEqual(
            msg.anymail_status.recipients["to1@example.com"].message_id,
            "aaaaaaaa-bbbb-cccc-dddd-eeeeeeeeeeee",
        )
        self.assertEqual(
            msg.anymail_status.esp_response.content, self.DEFAULT_RAW_RESPONSE
        )

    # noinspection PyUnresolvedReferences
    def test_send_failed_anymail_status(self):
        """If the send fails, anymail_status should contain initial values"""
        self.set_mock_response(status_code=500)
        sent = self.message.send(fail_silently=True)
        self.assertEqual(sent, 0)
        self.assertIsNone(self.message.anymail_status.status)
        self.assertIsNone(self.message.anymail_status.message_id)
        self.assertEqual(self.message.anymail_status.recipients, {})
        self.assertIsNone(self.message.anymail_status.esp_response)

    # noinspection PyUnresolvedReferences
    def test_send_unparsable_response(self):
        """
        If the send succeeds, but a non-JSON API response, should raise an API exception
        """
        mock_response = self.set_mock_response(
            status_code=200, raw=b"yikes, this isn't a real response"
        )
        with self.assertRaises(AnymailAPIError):
            self.message.send()
        self.assertIsNone(self.message.anymail_status.status)
        self.assertIsNone(self.message.anymail_status.message_id)
        self.assertEqual(self.message.anymail_status.recipients, {})
        self.assertEqual(self.message.anymail_status.esp_response, mock_response)

    def test_json_serialization_errors(self):
        """Try to provide more information about non-json-serializable data"""
        self.message.metadata = {"price": Decimal("19.99")}  # yeah, don't do this
        with self.assertRaises(AnymailSerializationError) as cm:
            self.message.send()
            print(self.get_api_call_json())
        err = cm.exception
        self.assertIsInstance(err, TypeError)  # compatibility with json.dumps
        # our added context:
        self.assertIn("Don't know how to send this data to Resend", str(err))
        # original message:
        self.assertRegex(str(err), r"Decimal.*is not JSON serializable")


@tag("resend")
class ResendBackendRecipientsRefusedTests(ResendBackendMockAPITestCase):
    # Resend doesn't check email bounce or complaint lists at time of send --
    # it always just queues the message. You'll need to listen for the "rejected"
    # and "failed" events to detect refused recipients.
    pass


@tag("resend")
class ResendBackendSessionSharingTestCase(
    SessionSharingTestCases, ResendBackendMockAPITestCase
):
    """Requests session sharing tests"""

    pass  # tests are defined in SessionSharingTestCases


@tag("resend")
@override_settings(EMAIL_BACKEND="anymail.backends.resend.EmailBackend")
class ResendBackendImproperlyConfiguredTests(AnymailTestMixin, SimpleTestCase):
    """Test ESP backend without required settings in place"""

    def test_missing_api_key(self):
        with self.assertRaises(ImproperlyConfigured) as cm:
            mail.send_mail("Subject", "Message", "from@example.com", ["to@example.com"])
        errmsg = str(cm.exception)
        self.assertRegex(errmsg, r"\bRESEND_API_KEY\b")
        self.assertRegex(errmsg, r"\bANYMAIL_RESEND_API_KEY\b")



from unittest.mock import ANY

from django.core.mail import EmailMessage, send_mail
from django.test import ignore_warnings, override_settings, tag

from anymail.exceptions import AnymailConfigurationError, AnymailDeprecationWarning
from anymail.webhooks.sendinblue import (
    SendinBlueInboundWebhookView,
    SendinBlueTrackingWebhookView,
)

from .mock_requests_backend import RequestsBackendMockAPITestCase
from .webhook_cases import WebhookTestCase


@tag("brevo", "sendinblue")
@override_settings(
    EMAIL_BACKEND="anymail.backends.sendinblue.EmailBackend",
    ANYMAIL={"SENDINBLUE_API_KEY": "test_api_key"},
)
@ignore_warnings(category=AnymailDeprecationWarning)
class SendinBlueBackendDeprecationTests(RequestsBackendMockAPITestCase):
    DEFAULT_RAW_RESPONSE = (
        b'{"messageId":"<201801020304.1234567890@smtp-relay.mailin.fr>"}'
    )
    DEFAULT_STATUS_CODE = 201  # Brevo v3 uses '201 Created' for success (in most cases)

    def test_deprecation_warning(self):
        message = EmailMessage(
            "Subject", "Body", "from@example.com", ["to@example.com"]
        )
        with self.assertWarnsMessage(
            AnymailDeprecationWarning,
            "`anymail.backends.sendinblue.EmailBackend` has been renamed"
            " `anymail.backends.brevo.EmailBackend`.",
        ):
            message.send()
        self.assert_esp_called("https://api.brevo.com/v3/smtp/email")

    @override_settings(ANYMAIL={"BREVO_API_KEY": "test_api_key"})
    def test_missing_api_key_error_uses_correct_setting_name(self):
        # The sendinblue.EmailBackend requires SENDINBLUE_ settings names
        with self.assertRaisesMessage(AnymailConfigurationError, "SENDINBLUE_API_KEY"):
            send_mail("Subject", "Body", "from@example.com", ["to@example.com"])


@tag("brevo", "sendinblue")
@ignore_warnings(category=AnymailDeprecationWarning)
class SendinBlueTrackingWebhookDeprecationTests(WebhookTestCase):
    def test_deprecation_warning(self):
        with self.assertWarnsMessage(
            AnymailDeprecationWarning,
            "Anymail's SendinBlue webhook URLs are deprecated.",
        ):
            response = self.client.post(
                "/anymail/sendinblue/tracking/",
                content_type="application/json",
                data="{}",
            )
        self.assertEqual(response.status_code, 200)
        # Old url uses old names to preserve compatibility:
        self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=SendinBlueTrackingWebhookView,  # *not* BrevoTrackingWebhookView
            event=ANY,
            esp_name="SendinBlue",  # *not* "Brevo"
        )

    def test_misconfigured_inbound(self):
        # Uses old esp_name when called on old URL
        errmsg = (
            "You seem to have set Brevo's *inbound* webhook URL"
            " to Anymail's SendinBlue *tracking* webhook URL."
        )
        with self.assertRaisesMessage(AnymailConfigurationError, errmsg):
            self.client.post(
                "/anymail/sendinblue/tracking/",
                content_type="application/json",
                data={"items": []},
            )


@tag("brevo", "sendinblue")
@override_settings(ANYMAIL_SENDINBLUE_API_KEY="test-api-key")
@ignore_warnings(category=AnymailDeprecationWarning)
class SendinBlueInboundWebhookDeprecationTests(WebhookTestCase):
    def test_deprecation_warning(self):
        with self.assertWarnsMessage(
            AnymailDeprecationWarning,
            "Anymail's SendinBlue webhook URLs are deprecated.",
        ):
            response = self.client.post(
                "/anymail/sendinblue/inbound/",
                content_type="application/json",
                data='{"items":[{}]}',
            )
        self.assertEqual(response.status_code, 200)
        # Old url uses old names to preserve compatibility:
        self.assert_handler_called_once_with(
            self.inbound_handler,
            sender=SendinBlueInboundWebhookView,  # *not* BrevoInboundWebhookView
            event=ANY,
            esp_name="SendinBlue",  # *not* "Brevo"
        )

    def test_misconfigured_tracking(self):
        # Uses old esp_name when called on old URL
        errmsg = (
            "You seem to have set Brevo's *tracking* webhook URL"
            " to Anymail's SendinBlue *inbound* webhook URL."
        )
        with self.assertRaisesMessage(AnymailConfigurationError, errmsg):
            self.client.post(
                "/anymail/sendinblue/inbound/",
                content_type="application/json",
                data={"event": "delivered"},
            )



from base64 import b64decode, b64encode
from calendar import timegm
from datetime import date, datetime
from decimal import Decimal
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage
from unittest.mock import patch

from django.core import mail
from django.test import SimpleTestCase, ignore_warnings, override_settings, tag
from django.utils.timezone import (
    get_fixed_timezone,
    override as override_current_timezone,
)

from anymail.exceptions import (
    AnymailAPIError,
    AnymailConfigurationError,
    AnymailNotSupportedWarning,
    AnymailSerializationError,
    AnymailUnsupportedFeature,
    AnymailWarning,
)
from anymail.message import attach_inline_image_file

from .mock_requests_backend import (
    RequestsBackendMockAPITestCase,
    SessionSharingTestCases,
)
from .utils import (
    SAMPLE_IMAGE_FILENAME,
    AnymailTestMixin,
    sample_image_content,
    sample_image_path,
)


@tag("sendgrid")
@override_settings(
    EMAIL_BACKEND="anymail.backends.sendgrid.EmailBackend",
    ANYMAIL={"SENDGRID_API_KEY": "test_api_key"},
)
@ignore_warnings(category=AnymailNotSupportedWarning)
class SendGridBackendMockAPITestCase(RequestsBackendMockAPITestCase):
    # SendGrid v3 success responses are empty:
    DEFAULT_RAW_RESPONSE = b""
    # SendGrid v3 uses '202 Accepted' for success (in most cases):
    DEFAULT_STATUS_CODE = 202

    def setUp(self):
        super().setUp()

        # Patch uuid4 to generate predictable anymail_ids for testing
        patch_uuid4 = patch(
            "anymail.backends.sendgrid.uuid.uuid4",
            side_effect=["mocked-uuid-%d" % n for n in range(1, 5)],
        )
        patch_uuid4.start()
        self.addCleanup(patch_uuid4.stop)

        # Simple message useful for many tests
        self.message = mail.EmailMultiAlternatives(
            "Subject", "Text Body", "from@example.com", ["to@example.com"]
        )


@tag("sendgrid")
@ignore_warnings(category=AnymailNotSupportedWarning)
class SendGridBackendStandardEmailTests(SendGridBackendMockAPITestCase):
    """Test backend support for Django standard email features"""

    def test_not_supported_warning(self):
        with self.assertWarns(
            AnymailNotSupportedWarning,
            msg="django-anymail has dropped official support for SendGrid.",
        ):
            mail.get_connection()

    def test_send_mail(self):
        """Test basic API for simple send"""
        mail.send_mail(
            "Subject here",
            "Here is the message.",
            "from@sender.example.com",
            ["to@example.com"],
            fail_silently=False,
        )
        self.assert_esp_called("https://api.sendgrid.com/v3/mail/send")
        http_headers = self.get_api_call_headers()
        self.assertEqual(http_headers["Authorization"], "Bearer test_api_key")
        self.assertEqual(http_headers["Content-Type"], "application/json")

        data = self.get_api_call_json()
        self.assertEqual(data["subject"], "Subject here")
        self.assertEqual(
            data["content"], [{"type": "text/plain", "value": "Here is the message."}]
        )
        self.assertEqual(data["from"], {"email": "from@sender.example.com"})
        self.assertEqual(
            data["personalizations"],
            [
                {
                    "to": [{"email": "to@example.com"}],
                    # make sure the backend assigned the anymail_id
                    # for event tracking and notification
                    "custom_args": {"anymail_id": "mocked-uuid-1"},
                }
            ],
        )

    def test_name_addr(self):
        """Make sure RFC2822 name-addr format (with display-name) is allowed

        (Test both sender and recipient addresses)
        """
        msg = mail.EmailMessage(
            "Subject",
            "Message",
            "From Name <from@example.com>",
            ["Recipient #1 <to1@example.com>", "to2@example.com"],
            cc=["Carbon Copy <cc1@example.com>", "cc2@example.com"],
            bcc=["Blind Copy <bcc1@example.com>", "bcc2@example.com"],
        )
        msg.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["from"], {"email": "from@example.com", "name": "From Name"}
        )

        # single message (single "personalization") sent to all those recipients
        self.assertEqual(len(data["personalizations"]), 1)
        self.assertEqual(
            data["personalizations"][0]["to"],
            [
                {"name": "Recipient #1", "email": "to1@example.com"},
                {"email": "to2@example.com"},
            ],
        )
        self.assertEqual(
            data["personalizations"][0]["cc"],
            [
                {"name": "Carbon Copy", "email": "cc1@example.com"},
                {"email": "cc2@example.com"},
            ],
        )
        self.assertEqual(
            data["personalizations"][0]["bcc"],
            [
                {"name": "Blind Copy", "email": "bcc1@example.com"},
                {"email": "bcc2@example.com"},
            ],
        )

    def test_email_message(self):
        email = mail.EmailMessage(
            "Subject",
            "Body goes here",
            "from@example.com",
            ["to1@example.com", "Also To <to2@example.com>"],
            bcc=["bcc1@example.com", "Also BCC <bcc2@example.com>"],
            cc=["cc1@example.com", "Also CC <cc2@example.com>"],
            headers={
                "Reply-To": "another@example.com",
                "X-MyHeader": "my value",
                # should override backend msgid:
                "Message-ID": "<mycustommsgid@sales.example.com>",
            },
        )
        email.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["personalizations"],
            [
                {
                    "to": [
                        {"email": "to1@example.com"},
                        {"email": "to2@example.com", "name": "Also To"},
                    ],
                    "cc": [
                        {"email": "cc1@example.com"},
                        {"email": "cc2@example.com", "name": "Also CC"},
                    ],
                    "bcc": [
                        {"email": "bcc1@example.com"},
                        {"email": "bcc2@example.com", "name": "Also BCC"},
                    ],
                    # make sure custom Message-ID also added to custom_args
                    "custom_args": {"anymail_id": "mocked-uuid-1"},
                }
            ],
        )

        self.assertEqual(data["from"], {"email": "from@example.com"})
        self.assertEqual(data["subject"], "Subject")
        self.assertEqual(
            data["content"], [{"type": "text/plain", "value": "Body goes here"}]
        )
        self.assertEqual(data["reply_to_list"], [{"email": "another@example.com"}])
        self.assertEqual(
            data["headers"],
            {
                "X-MyHeader": "my value",
                "Message-ID": "<mycustommsgid@sales.example.com>",
            },
        )

    def test_html_message(self):
        text_content = "This is an important message."
        html_content = "<p>This is an <strong>important</strong> message.</p>"
        email = mail.EmailMultiAlternatives(
            "Subject", text_content, "from@example.com", ["to@example.com"]
        )
        email.attach_alternative(html_content, "text/html")
        email.send()
        data = self.get_api_call_json()
        # SendGrid requires content in text, html order:
        self.assertEqual(len(data["content"]), 2)
        self.assertEqual(
            data["content"][0], {"type": "text/plain", "value": text_content}
        )
        self.assertEqual(
            data["content"][1], {"type": "text/html", "value": html_content}
        )
        # Don't accidentally send the html part as an attachment:
        self.assertNotIn("attachments", data)

    def test_html_only_message(self):
        html_content = "<p>This is an <strong>important</strong> message.</p>"
        email = mail.EmailMessage(
            "Subject", html_content, "from@example.com", ["to@example.com"]
        )
        email.content_subtype = "html"  # Main content is now text/html
        email.send()
        data = self.get_api_call_json()
        self.assertEqual(len(data["content"]), 1)
        self.assertEqual(
            data["content"][0], {"type": "text/html", "value": html_content}
        )

    def test_extra_headers(self):
        self.message.extra_headers = {
            "X-Custom": "string",
            "X-Num": 123,
            "Reply-To": '"Do Not Reply" <noreply@example.com>',
        }
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["headers"]["X-Custom"], "string")
        # number converted to string (undocumented SendGrid requirement):
        self.assertEqual(data["headers"]["X-Num"], "123")
        # Reply-To must be moved to separate param
        self.assertNotIn("Reply-To", data["headers"])
        self.assertEqual(
            data["reply_to_list"],
            [{"name": "Do Not Reply", "email": "noreply@example.com"}],
        )

    def test_extra_headers_serialization_error(self):
        self.message.extra_headers = {"X-Custom": Decimal(12.5)}
        with self.assertRaisesMessage(AnymailSerializationError, "Decimal"):
            self.message.send()

    def test_reply_to(self):
        self.message.reply_to = [
            '"Reply recipient" <reply@example.com',
            "reply2@example.com",
        ]
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["reply_to_list"],
            [
                {"name": "Reply recipient", "email": "reply@example.com"},
                {"email": "reply2@example.com"},
            ],
        )
        self.assertNotIn("reply_to", data)  # not allowed with reply_to_list

    def test_attachments(self):
        text_content = "* Item one\n* Item two\n* Item three"
        self.message.attach(
            filename="test.txt", content=text_content, mimetype="text/plain"
        )

        # Should guess mimetype if not provided...
        png_content = b"PNG\xb4 pretend this is the contents of a png file"
        self.message.attach(filename="test.png", content=png_content)

        # Should work with a MIMEBase object (also tests no filename)...
        pdf_content = b"PDF\xb4 pretend this is valid pdf data"
        mimeattachment = MIMEBase("application", "pdf")
        mimeattachment.set_payload(pdf_content)
        self.message.attach(mimeattachment)

        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(len(data["attachments"]), 3)

        attachments = data["attachments"]
        self.assertEqual(
            attachments[0],
            {
                "filename": "test.txt",
                "content": b64encode(text_content.encode("utf-8")).decode("ascii"),
                "type": "text/plain",
            },
        )
        self.assertEqual(
            attachments[1],
            {
                "filename": "test.png",
                "content": b64encode(png_content).decode("ascii"),
                "type": "image/png",  # (type inferred from filename)
            },
        )
        self.assertEqual(
            attachments[2],
            {
                "filename": "",  # no filename -- but param is required
                "content": b64encode(pdf_content).decode("ascii"),
                "type": "application/pdf",
            },
        )

    def test_unicode_attachment_correctly_decoded(self):
        self.message.attach(
            "Une pièce jointe.html", "<p>\u2019</p>", mimetype="text/html"
        )
        self.message.send()
        attachment = self.get_api_call_json()["attachments"][0]
        self.assertEqual(attachment["filename"], "Une pièce jointe.html")
        self.assertEqual(
            b64decode(attachment["content"]).decode("utf-8"), "<p>\u2019</p>"
        )

    def test_embedded_images(self):
        image_filename = SAMPLE_IMAGE_FILENAME
        image_path = sample_image_path(image_filename)
        image_data = sample_image_content(image_filename)

        cid = attach_inline_image_file(self.message, image_path)  # Read from a png file
        html_content = (
            '<p>This has an <img src="cid:%s" alt="inline" /> image.</p>' % cid
        )
        self.message.attach_alternative(html_content, "text/html")

        self.message.send()
        data = self.get_api_call_json()

        self.assertEqual(
            data["attachments"][0],
            {
                "filename": image_filename,
                "content": b64encode(image_data).decode("ascii"),
                "type": "image/png",  # (type inferred from filename)
                "disposition": "inline",
                "content_id": cid,
            },
        )

    def test_attached_images(self):
        image_filename = SAMPLE_IMAGE_FILENAME
        image_path = sample_image_path(image_filename)
        image_data = sample_image_content(image_filename)

        # option 1: attach as a file
        self.message.attach_file(image_path)

        # option 2: construct the MIMEImage and attach it directly
        image = MIMEImage(image_data)
        self.message.attach(image)

        self.message.send()

        image_data_b64 = b64encode(image_data).decode("ascii")
        data = self.get_api_call_json()
        self.assertEqual(
            data["attachments"][0],
            {
                "filename": image_filename,  # the named one
                "content": image_data_b64,
                "type": "image/png",
            },
        )
        self.assertEqual(
            data["attachments"][1],
            {
                "filename": "",  # the unnamed one
                "content": image_data_b64,
                "type": "image/png",
            },
        )

    def test_multiple_html_alternatives(self):
        # SendGrid's v3 API allows all kinds of content alternatives.
        # It's unclear whether this would permit multiple text/html parts
        # (the API docs warn that "If included, text/plain and text/html must be
        # the first indices of the [content] array in this order"), but Anymail
        # generally passes whatever the API structure supports -- deferring any
        # limitations to the ESP.
        self.message.body = "Text body"
        self.message.attach_alternative("<p>First html is OK</p>", "text/html")
        self.message.attach_alternative(
            "<p>And maybe second html, too</p>", "text/html"
        )

        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["content"],
            [
                {"type": "text/plain", "value": "Text body"},
                {"type": "text/html", "value": "<p>First html is OK</p>"},
                {"type": "text/html", "value": "<p>And maybe second html, too</p>"},
            ],
        )

    def test_non_html_alternative(self):
        self.message.body = "Text body"
        self.message.attach_alternative("{'maybe': 'allowed'}", "application/json")
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["content"],
            [
                {"type": "text/plain", "value": "Text body"},
                {"type": "application/json", "value": "{'maybe': 'allowed'}"},
            ],
        )

    def test_api_failure(self):
        self.set_mock_response(status_code=400)
        with self.assertRaisesMessage(AnymailAPIError, "SendGrid API response 400"):
            mail.send_mail("Subject", "Body", "from@example.com", ["to@example.com"])

        # Make sure fail_silently is respected
        self.set_mock_response(status_code=400)
        sent = mail.send_mail(
            "Subject",
            "Body",
            "from@example.com",
            ["to@example.com"],
            fail_silently=True,
        )
        self.assertEqual(sent, 0)

    def test_api_error_includes_details(self):
        """AnymailAPIError should include ESP's error message"""
        # JSON error response:
        error_response = b"""{"errors":[
            {"message":"Helpful explanation from SendGrid",
             "field":"subject","help":null},
            {"message":"Another error","field":null,"help":null}
        ]}"""
        self.set_mock_response(status_code=400, raw=error_response)
        with self.assertRaises(AnymailAPIError) as cm:
            self.message.send()
        err = cm.exception
        self.assertIn("Helpful explanation from SendGrid", str(err))
        self.assertIn("Another error", str(err))

        # Non-JSON error response:
        self.set_mock_response(status_code=500, raw=b"Ack! Bad proxy!")
        with self.assertRaisesMessage(AnymailAPIError, "Ack! Bad proxy!"):
            self.message.send()

        # No content in the error response:
        self.set_mock_response(status_code=502, raw=None)
        with self.assertRaises(AnymailAPIError):
            self.message.send()


@tag("sendgrid")
@ignore_warnings(category=AnymailNotSupportedWarning)
class SendGridBackendAnymailFeatureTests(SendGridBackendMockAPITestCase):
    """Test backend support for Anymail added features"""

    def test_envelope_sender(self):
        # SendGrid does not have a way to change envelope sender.
        self.message.envelope_sender = "anything@bounces.example.com"
        with self.assertRaisesMessage(AnymailUnsupportedFeature, "envelope_sender"):
            self.message.send()

    def test_metadata(self):
        self.message.metadata = {"user_id": "12345", "items": 6, "float": 98.6}
        self.message.send()
        data = self.get_api_call_json()
        # remove anymail_id we added for tracking:
        data["custom_args"].pop("anymail_id", None)
        self.assertEqual(
            data["custom_args"],
            {
                "user_id": "12345",
                "items": "6",  # int converted to a string,
                "float": "98.6",  # float converted to a string (watch binary rounding!)
            },
        )

    def test_send_at(self):
        utc_plus_6 = get_fixed_timezone(6 * 60)
        utc_minus_8 = get_fixed_timezone(-8 * 60)

        with override_current_timezone(utc_plus_6):
            # Timezone-aware datetime converted to UTC:
            self.message.send_at = datetime(2016, 3, 4, 5, 6, 7, tzinfo=utc_minus_8)
            self.message.send()
            data = self.get_api_call_json()
            self.assertEqual(
                data["send_at"], timegm((2016, 3, 4, 13, 6, 7))
            )  # 05:06 UTC-8 == 13:06 UTC

            # Timezone-naive datetime assumed to be Django current_timezone
            self.message.send_at = datetime(
                2022, 10, 11, 12, 13, 14, 567
            )  # microseconds should get stripped
            self.message.send()
            data = self.get_api_call_json()
            self.assertEqual(
                data["send_at"], timegm((2022, 10, 11, 6, 13, 14))
            )  # 12:13 UTC+6 == 06:13 UTC

            # Date-only treated as midnight in current timezone
            self.message.send_at = date(2022, 10, 22)
            self.message.send()
            data = self.get_api_call_json()
            self.assertEqual(
                data["send_at"], timegm((2022, 10, 21, 18, 0, 0))
            )  # 00:00 UTC+6 == 18:00-1d UTC

            # POSIX timestamp
            self.message.send_at = 1651820889  # 2022-05-06 07:08:09 UTC
            self.message.send()
            data = self.get_api_call_json()
            self.assertEqual(data["send_at"], 1651820889)

    def test_tags(self):
        self.message.tags = ["receipt", "repeat-user"]
        self.message.send()
        data = self.get_api_call_json()
        self.assertCountEqual(data["categories"], ["receipt", "repeat-user"])

    def test_tracking(self):
        # Test one way...
        self.message.track_clicks = False
        self.message.track_opens = True
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["tracking_settings"]["click_tracking"], {"enable": False})
        self.assertEqual(data["tracking_settings"]["open_tracking"], {"enable": True})

        # ...and the opposite way
        self.message.track_clicks = True
        self.message.track_opens = False
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["tracking_settings"]["click_tracking"], {"enable": True})
        self.assertEqual(data["tracking_settings"]["open_tracking"], {"enable": False})

    def test_template_id(self):
        self.message.template_id = "5997fcf6-2b9f-484d-acd5-7e9a99f0dc1f"
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(data["template_id"], "5997fcf6-2b9f-484d-acd5-7e9a99f0dc1f")

    def test_template_id_with_empty_body(self):
        # v2 API required *some* text and html in message to render those template
        # bodies, so the v2 backend set those to " " when necessary.
        # But per v3 docs:
        #   "If you use a template that contains content and a subject (either text or
        #   html), you do not need to specify those in the respective personalizations
        #   or message level parameters."
        # So make sure we aren't adding body content where not needed:
        message = mail.EmailMessage(
            from_email="from@example.com", to=["to@example.com"]
        )
        message.template_id = "5997fcf6-2b9f-484d-acd5-7e9a99f0dc1f"
        message.send()
        data = self.get_api_call_json()
        self.assertNotIn("content", data)  # neither text nor html body
        self.assertNotIn("subject", data)

    def test_merge_data(self):
        # A template_id starting with "d-" indicates you are using SendGrid's newer
        # (non-legacy) "dynamic" transactional templates
        self.message.template_id = "d-5a963add2ec84305813ff860db277d7a"

        self.message.from_email = "from@example.com"
        self.message.to = [
            "alice@example.com",
            "Bob <bob@example.com>",
            "celia@example.com",
        ]
        # cc gets applied to *each* recipient in a merge:
        self.message.cc = ["cc@example.com"]

        self.message.merge_data = {
            "alice@example.com": {"name": "Alice", "group": "Developers"},
            "bob@example.com": {"name": "Bob"},  # and leave group undefined
            # and no data for celia@example.com
        }
        self.message.merge_global_data = {
            "group": "Users",
            "site": "ExampleCo",
        }
        self.message.send()

        data = self.get_api_call_json()
        self.assertEqual(
            data["personalizations"],
            [
                {
                    "to": [{"email": "alice@example.com"}],
                    "cc": [{"email": "cc@example.com"}],  # all recipients get the cc
                    "custom_args": {"anymail_id": "mocked-uuid-1"},
                    "dynamic_template_data": {
                        "name": "Alice",
                        "group": "Developers",
                        "site": "ExampleCo",
                    },
                },
                {
                    "to": [{"email": "bob@example.com", "name": "Bob"}],
                    "cc": [{"email": "cc@example.com"}],
                    "custom_args": {"anymail_id": "mocked-uuid-2"},
                    "dynamic_template_data": {
                        "name": "Bob",
                        "group": "Users",
                        "site": "ExampleCo",
                    },
                },
                {
                    "to": [{"email": "celia@example.com"}],
                    "cc": [{"email": "cc@example.com"}],
                    "custom_args": {"anymail_id": "mocked-uuid-3"},
                    "dynamic_template_data": {"group": "Users", "site": "ExampleCo"},
                },
            ],
        )
        self.assertNotIn("sections", data)  # 'sections' not used with dynamic templates

    def test_explicit_dynamic_template(self):
        # undocumented esp_extra['use_dynamic_template']
        # can be used to force dynamic/legacy params
        self.message.merge_data = {"to@example.com": {"test": "data"}}

        self.message.template_id = "apparently-not-dynamic"  # doesn't start with "d-"
        self.message.esp_extra = {"use_dynamic_template": True}
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["personalizations"],
            [
                {
                    "to": [{"email": "to@example.com"}],
                    "custom_args": {"anymail_id": "mocked-uuid-1"},
                    "dynamic_template_data": {"test": "data"},
                }
            ],
        )

        self.message.template_id = "d-apparently-not-legacy"
        self.message.esp_extra = {
            "use_dynamic_template": False,
            "merge_field_format": "<%{}%>",
        }
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["personalizations"],
            [
                {
                    "to": [{"email": "to@example.com"}],
                    "custom_args": {"anymail_id": "mocked-uuid-2"},
                    "substitutions": {"<%test%>": "data"},
                }
            ],
        )

    def test_merge_data_global_only(self):
        # a template with only global data can be used to send the same message
        # to multiple recipients (non-batch)
        self.message.template_id = "d-5a963add2ec84305813ff860db277d7a"
        self.message.merge_global_data = {"test": "data"}
        self.message.to = ["one@example.com", "two@example.com"]
        self.message.send()

        data = self.get_api_call_json()
        self.assertEqual(
            data["personalizations"],
            [
                {
                    "to": [  # not batch
                        {"email": "one@example.com"},
                        {"email": "two@example.com"},
                    ],
                    "custom_args": {"anymail_id": "mocked-uuid-1"},
                    "dynamic_template_data": {"test": "data"},
                }
            ],
        )

    def test_legacy_merge_data(self):
        # unless a new "dynamic template" is specified, Anymail assumes the legacy
        # "substitutions" format for merge data
        self.message.from_email = "from@example.com"
        self.message.to = [
            "alice@example.com",
            "Bob <bob@example.com>",
            "celia@example.com",
        ]
        # cc gets applied to *each* recipient in a merge:
        self.message.cc = ["cc@example.com"]
        # SendGrid template_id is not required to use merge.
        # You can just supply (legacy) template content as the message (e.g.):
        self.message.body = "Hi :name. Welcome to :group at :site."
        self.message.merge_data = {
            # You must either include merge field delimiters in the keys
            # (':name' rather than just 'name') as shown here, or use one of the
            # merge_field_format options shown in the test cases below
            "alice@example.com": {":name": "Alice", ":group": "Developers"},
            "bob@example.com": {":name": "Bob"},  # and leave :group undefined
            # and no data for celia@example.com
        }
        self.message.merge_global_data = {
            ":group": "Users",
            ":site": "ExampleCo",
        }
        self.message.send()

        data = self.get_api_call_json()
        self.assertEqual(
            data["personalizations"],
            [
                {
                    "to": [{"email": "alice@example.com"}],
                    "cc": [{"email": "cc@example.com"}],  # all recipients get the cc
                    "custom_args": {"anymail_id": "mocked-uuid-1"},
                    "substitutions": {
                        ":name": "Alice",
                        ":group": "Developers",
                        ":site": "ExampleCo",  # merge_global_data merged
                    },
                },
                {
                    "to": [{"email": "bob@example.com", "name": "Bob"}],
                    "cc": [{"email": "cc@example.com"}],
                    "custom_args": {"anymail_id": "mocked-uuid-2"},
                    "substitutions": {
                        ":name": "Bob",
                        ":group": "Users",
                        ":site": "ExampleCo",
                    },
                },
                {
                    "to": [{"email": "celia@example.com"}],
                    "cc": [{"email": "cc@example.com"}],
                    "custom_args": {"anymail_id": "mocked-uuid-3"},
                    "substitutions": {":group": "Users", ":site": "ExampleCo"},
                },
            ],
        )
        # 'sections' no longer used for merge_global_data:
        self.assertNotIn("sections", data)

    @override_settings(
        ANYMAIL_SENDGRID_MERGE_FIELD_FORMAT=":{}"  # :field as shown in SG examples
    )
    def test_legacy_merge_field_format_setting(self):
        # Provide merge field delimiters in settings.py
        self.message.to = ["alice@example.com", "Bob <bob@example.com>"]
        self.message.merge_data = {
            "alice@example.com": {"name": "Alice", "group": "Developers"},
            "bob@example.com": {"name": "Bob"},  # and leave group undefined
        }
        self.message.merge_global_data = {"site": "ExampleCo"}
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["personalizations"],
            [
                {
                    "to": [{"email": "alice@example.com"}],
                    "custom_args": {"anymail_id": "mocked-uuid-1"},
                    "substitutions": {
                        ":name": "Alice",
                        ":group": "Developers",  # keys changed to :field
                        ":site": "ExampleCo",
                    },
                },
                {
                    "to": [{"email": "bob@example.com", "name": "Bob"}],
                    "custom_args": {"anymail_id": "mocked-uuid-2"},
                    "substitutions": {":name": "Bob", ":site": "ExampleCo"},
                },
            ],
        )

    def test_legacy_merge_field_format_esp_extra(self):
        # Provide merge field delimiters for an individual message
        self.message.to = ["alice@example.com", "Bob <bob@example.com>"]
        self.message.merge_data = {
            "alice@example.com": {"name": "Alice", "group": "Developers"},
            "bob@example.com": {"name": "Bob"},  # and leave group undefined
        }
        self.message.merge_global_data = {"site": "ExampleCo"}
        # match Mandrill/MailChimp *|field|* delimiters:
        self.message.esp_extra = {"merge_field_format": "*|{}|*"}
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["personalizations"],
            [
                {
                    "to": [{"email": "alice@example.com"}],
                    "custom_args": {"anymail_id": "mocked-uuid-1"},
                    "substitutions": {
                        "*|name|*": "Alice",
                        "*|group|*": "Developers",
                        "*|site|*": "ExampleCo",
                    },
                },
                {
                    "to": [{"email": "bob@example.com", "name": "Bob"}],
                    "custom_args": {"anymail_id": "mocked-uuid-2"},
                    "substitutions": {"*|name|*": "Bob", "*|site|*": "ExampleCo"},
                },
            ],
        )
        # Make sure our esp_extra merge_field_format doesn't get sent to SendGrid API:
        self.assertNotIn("merge_field_format", data)

    def test_legacy_warn_if_no_merge_field_delimiters(self):
        self.message.to = ["alice@example.com"]
        self.message.merge_data = {
            "alice@example.com": {"name": "Alice", "group": "Developers"},
        }
        with self.assertWarnsRegex(AnymailWarning, r"SENDGRID_MERGE_FIELD_FORMAT"):
            self.message.send()

    def test_legacy_warn_if_no_global_merge_field_delimiters(self):
        self.message.merge_global_data = {"site": "ExampleCo"}
        with self.assertWarnsRegex(AnymailWarning, r"SENDGRID_MERGE_FIELD_FORMAT"):
            self.message.send()

    def test_legacy_data_conversion(self):
        """
        SendGrid requires string (or null) substitution values.
        Anymail will convert numbers.
        """
        # (Legacy substitutions path because not using dynamic template id.)
        self.message.to = ["alice@example.com"]
        self.message.merge_data = {
            "alice@example.com": {
                ":integer": 1,
                ":float": 1.0,
                ":null": None,
            }
        }
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["personalizations"][0]["substitutions"],
            {
                ":integer": "1",
                ":float": "1.0",
                ":null": None,
            },
        )

    def test_legacy_unsupported_string_conversion(self):
        """
        SendGrid requires string substitution values (and issues a cryptic
        error otherwise). Anymail treats non-convertible data as an error.
        """
        # (Legacy substitutions path because not using dynamic template id.)
        self.message.to = ["alice@example.com"]
        cases = [
            ["array"],
            {"dict": 1},
            Decimal("1.0"),
        ]
        for case in cases:
            with self.subTest(case=case):
                self.message.merge_data = {
                    "alice@example.com": {
                        ":value": case,
                    }
                }
                with self.assertRaisesMessage(
                    AnymailSerializationError,
                    "SendGrid legacy substitutions require string values",
                ):
                    self.message.send()

    def test_merge_metadata(self):
        self.message.to = ["alice@example.com", "Bob <bob@example.com>"]
        self.message.merge_metadata = {
            "alice@example.com": {"order_id": 123},
            "bob@example.com": {"order_id": 678, "tier": "premium"},
        }
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["personalizations"],
            [
                {
                    "to": [{"email": "alice@example.com"}],
                    # anymail_id added to other custom_args
                    "custom_args": {"anymail_id": "mocked-uuid-1", "order_id": "123"},
                },
                {
                    "to": [{"email": "bob@example.com", "name": "Bob"}],
                    "custom_args": {
                        "anymail_id": "mocked-uuid-2",
                        "order_id": "678",
                        "tier": "premium",
                    },
                },
            ],
        )

    def test_metadata_with_merge_metadata(self):
        # Per SendGrid docs: "personalizations[x].custom_args will be merged
        # with message level custom_args, overriding any conflicting keys."
        # So there's no need to merge global metadata with per-recipient merge_metadata
        # (like we have to for template merge_global_data and merge_data).
        self.message.to = ["alice@example.com", "Bob <bob@example.com>"]
        self.message.metadata = {"tier": "basic", "batch": "ax24"}
        self.message.merge_metadata = {
            "alice@example.com": {"order_id": 123},
            "bob@example.com": {"order_id": 678, "tier": "premium"},
        }
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["personalizations"],
            [
                {
                    "to": [{"email": "alice@example.com"}],
                    "custom_args": {"anymail_id": "mocked-uuid-1", "order_id": "123"},
                },
                {
                    "to": [{"email": "bob@example.com", "name": "Bob"}],
                    "custom_args": {
                        "anymail_id": "mocked-uuid-2",
                        "order_id": "678",
                        "tier": "premium",
                    },
                },
            ],
        )
        self.assertEqual(data["custom_args"], {"tier": "basic", "batch": "ax24"})

    def test_merge_metadata_with_merge_data(self):
        # (using dynamic templates)
        self.message.to = [
            "alice@example.com",
            "Bob <bob@example.com>",
            "celia@example.com",
        ]
        # cc gets applied to *each* recipient in a merge:
        self.message.cc = ["cc@example.com"]
        self.message.template_id = "d-5a963add2ec84305813ff860db277d7a"
        self.message.merge_data = {
            "alice@example.com": {"name": "Alice", "group": "Developers"},
            "bob@example.com": {"name": "Bob"},
            # and no data for celia@example.com
        }
        self.message.merge_global_data = {
            "group": "Users",
            "site": "ExampleCo",
        }
        self.message.merge_metadata = {
            "alice@example.com": {"order_id": 123},
            "bob@example.com": {"order_id": 678, "tier": "premium"},
            # and no metadata for celia@example.com
        }
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["personalizations"],
            [
                {
                    "to": [{"email": "alice@example.com"}],
                    "cc": [{"email": "cc@example.com"}],  # all recipients get the cc
                    "dynamic_template_data": {
                        "name": "Alice",
                        "group": "Developers",
                        "site": "ExampleCo",
                    },
                    "custom_args": {"anymail_id": "mocked-uuid-1", "order_id": "123"},
                },
                {
                    "to": [{"email": "bob@example.com", "name": "Bob"}],
                    "cc": [{"email": "cc@example.com"}],
                    "dynamic_template_data": {
                        "name": "Bob",
                        "group": "Users",
                        "site": "ExampleCo",
                    },
                    "custom_args": {
                        "anymail_id": "mocked-uuid-2",
                        "order_id": "678",
                        "tier": "premium",
                    },
                },
                {
                    "to": [{"email": "celia@example.com"}],
                    "cc": [{"email": "cc@example.com"}],
                    "dynamic_template_data": {"group": "Users", "site": "ExampleCo"},
                    "custom_args": {"anymail_id": "mocked-uuid-3"},
                },
            ],
        )

    def test_merge_metadata_with_legacy_template(self):
        self.message.to = [
            "alice@example.com",
            "Bob <bob@example.com>",
            "celia@example.com",
        ]
        # cc gets applied to *each* recipient in a merge:
        self.message.cc = ["cc@example.com"]
        self.message.template_id = "5a963add2ec84305813ff860db277d7a"
        self.message.esp_extra = {"merge_field_format": ":{}"}
        self.message.merge_data = {
            "alice@example.com": {"name": "Alice", "group": "Developers"},
            "bob@example.com": {"name": "Bob"},
            # and no data for celia@example.com
        }
        self.message.merge_global_data = {
            "group": "Users",
            "site": "ExampleCo",
        }
        self.message.merge_metadata = {
            "alice@example.com": {"order_id": 123},
            "bob@example.com": {"order_id": 678, "tier": "premium"},
            # and no metadata for celia@example.com
        }
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["personalizations"],
            [
                {
                    "to": [{"email": "alice@example.com"}],
                    "cc": [{"email": "cc@example.com"}],  # all recipients get the cc
                    "custom_args": {"anymail_id": "mocked-uuid-1", "order_id": "123"},
                    "substitutions": {
                        ":name": "Alice",
                        ":group": "Developers",
                        ":site": "ExampleCo",
                    },
                },
                {
                    "to": [{"email": "bob@example.com", "name": "Bob"}],
                    "cc": [{"email": "cc@example.com"}],
                    "custom_args": {
                        "anymail_id": "mocked-uuid-2",
                        "order_id": "678",
                        "tier": "premium",
                    },
                    "substitutions": {
                        ":name": "Bob",
                        ":group": "Users",
                        ":site": "ExampleCo",
                    },
                },
                {
                    "to": [{"email": "celia@example.com"}],
                    "cc": [{"email": "cc@example.com"}],
                    "custom_args": {"anymail_id": "mocked-uuid-3"},
                    "substitutions": {":group": "Users", ":site": "ExampleCo"},
                },
            ],
        )

    def test_merge_headers(self):
        self.message.to = ["alice@example.com", "Bob <bob@example.com>"]
        self.message.extra_headers = {
            "List-Unsubscribe-Post": "List-Unsubscribe=One-Click",
            "List-Unsubscribe": "<mailto:unsubscribe@example.com>",
        }
        self.message.merge_headers = {
            "alice@example.com": {
                "List-Unsubscribe": "<https://example.com/a/>",
            },
            "bob@example.com": {
                "List-Unsubscribe": "<https://example.com/b/>",
            },
        }
        self.message.send()
        data = self.get_api_call_json()
        personalizations = data["personalizations"]
        self.assertEqual(len(personalizations), 2)

        self.assertEqual(personalizations[0]["to"][0]["email"], "alice@example.com")
        self.assertEqual(
            personalizations[0]["headers"],
            {"List-Unsubscribe": "<https://example.com/a/>"},
        )
        self.assertEqual(personalizations[1]["to"][0]["email"], "bob@example.com")
        self.assertEqual(
            personalizations[1]["headers"],
            {"List-Unsubscribe": "<https://example.com/b/>"},
        )

        # non-merge headers still in globals:
        self.assertEqual(
            data["headers"],
            {
                "List-Unsubscribe-Post": "List-Unsubscribe=One-Click",
                "List-Unsubscribe": "<mailto:unsubscribe@example.com>",
            },
        )

    @override_settings(
        ANYMAIL_SENDGRID_GENERATE_MESSAGE_ID=False  # else we force custom_args
    )
    def test_default_omits_options(self):
        """Make sure by default we don't send any ESP-specific options.

        Options not specified by the caller should be omitted entirely from
        the API call (*not* sent as False or empty). This ensures
        that your ESP account settings apply by default.
        """
        self.message.send()
        data = self.get_api_call_json()
        self.assertNotIn("asm", data)
        self.assertNotIn("attachments", data)
        self.assertNotIn("batch_id", data)
        self.assertNotIn("categories", data)
        self.assertNotIn("custom_args", data)
        self.assertNotIn("headers", data)
        self.assertNotIn("ip_pool_name", data)
        self.assertNotIn("mail_settings", data)
        self.assertNotIn("reply_to", data)
        self.assertNotIn("reply_to_list", data)
        self.assertNotIn("sections", data)
        self.assertNotIn("send_at", data)
        self.assertNotIn("template_id", data)
        self.assertNotIn("tracking_settings", data)

        for personalization in data["personalizations"]:
            self.assertNotIn("custom_args", personalization)
            self.assertNotIn("dynamic_template_data", personalization)
            self.assertNotIn("headers", personalization)
            self.assertNotIn("send_at", personalization)
            self.assertNotIn("substitutions", personalization)

    def test_esp_extra(self):
        self.message.tags = ["tag"]
        self.message.track_clicks = True
        self.message.esp_extra = {
            "ip_pool_name": "transactional",
            "asm": {  # subscription management
                "group_id": 1,
            },
            "tracking_settings": {
                "subscription_tracking": {
                    "enable": True,
                    "substitution_tag": "[unsubscribe_url]",
                },
            },
        }
        self.message.send()
        data = self.get_api_call_json()
        # merged from esp_extra:
        self.assertEqual(data["ip_pool_name"], "transactional")
        self.assertEqual(data["asm"], {"group_id": 1})
        self.assertEqual(
            data["tracking_settings"]["subscription_tracking"],
            {"enable": True, "substitution_tag": "[unsubscribe_url]"},
        )
        # make sure we didn't overwrite Anymail message options:
        self.assertEqual(data["categories"], ["tag"])
        self.assertEqual(data["tracking_settings"]["click_tracking"], {"enable": True})

    def test_esp_extra_pesonalizations(self):
        self.message.to = ["First recipient <first@example.com>", "second@example.com"]
        self.message.merge_data = {}  # force separate messages for each 'to'

        # esp_extra['personalizations'] dict
        # merges with message-derived personalizations
        self.message.esp_extra = {"personalizations": {"future_feature": "works"}}
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["personalizations"],
            [
                {
                    "to": [{"email": "first@example.com", "name": "First recipient"}],
                    "custom_args": {"anymail_id": "mocked-uuid-1"},
                    "future_feature": "works",
                },
                {
                    "to": [{"email": "second@example.com"}],
                    "custom_args": {"anymail_id": "mocked-uuid-2"},
                    "future_feature": "works",  # merged into *every* recipient
                },
            ],
        )

        # but esp_extra['personalizations'] list just overrides entire personalizations
        # (for backwards compatibility)
        self.message.esp_extra = {
            "personalizations": [
                {"to": [{"email": "custom@example.com"}], "future_feature": "works"}
            ]
        }
        self.message.send()
        data = self.get_api_call_json()
        self.assertEqual(
            data["personalizations"],
            [
                {
                    "to": [{"email": "custom@example.com"}],
                    "custom_args": {"anymail_id": "mocked-uuid-3"},
                    "future_feature": "works",
                },
            ],
        )

    # noinspection PyUnresolvedReferences
    def test_send_attaches_anymail_status(self):
        """The anymail_status should be attached to the message when it is sent"""
        # the DEFAULT_RAW_RESPONSE above is the *only* success response SendGrid
        # returns, so no need to override it here
        msg = mail.EmailMessage(
            "Subject",
            "Message",
            "from@example.com",
            ["to1@example.com"],
        )
        sent = msg.send()
        self.assertEqual(sent, 1)
        self.assertEqual(msg.anymail_status.status, {"queued"})
        self.assertEqual(msg.anymail_status.message_id, "mocked-uuid-1")
        self.assertEqual(
            msg.anymail_status.recipients["to1@example.com"].status, "queued"
        )
        self.assertEqual(
            msg.anymail_status.recipients["to1@example.com"].message_id, "mocked-uuid-1"
        )
        self.assertEqual(
            msg.anymail_status.esp_response.content, self.DEFAULT_RAW_RESPONSE
        )

    def test_batch_recipients_get_unique_message_ids(self):
        """In a batch send, each recipient should get a distinct own message_id"""
        msg = mail.EmailMessage(
            "Subject",
            "Message",
            "from@example.com",
            ["to1@example.com", "Someone Else <to2@example.com>"],
            cc=["cc@example.com"],
        )
        msg.merge_data = {}  # force batch send
        msg.send()
        self.assertEqual(
            msg.anymail_status.message_id, {"mocked-uuid-1", "mocked-uuid-2"}
        )
        self.assertEqual(
            msg.anymail_status.recipients["to1@example.com"].message_id, "mocked-uuid-1"
        )
        self.assertEqual(
            msg.anymail_status.recipients["to2@example.com"].message_id, "mocked-uuid-2"
        )
        # cc's (and bcc's) get copied for all batch recipients,
        # but we can only communicate one message_id:
        self.assertEqual(
            msg.anymail_status.recipients["cc@example.com"].message_id, "mocked-uuid-2"
        )

    @override_settings(ANYMAIL_SENDGRID_GENERATE_MESSAGE_ID=False)
    def test_disable_generate_message_id(self):
        msg = mail.EmailMessage(
            "Subject",
            "Message",
            "from@example.com",
            ["to1@example.com"],
        )
        msg.send()
        self.assertIsNone(msg.anymail_status.message_id)
        self.assertIsNone(msg.anymail_status.recipients["to1@example.com"].message_id)

    # noinspection PyUnresolvedReferences
    def test_send_failed_anymail_status(self):
        """If the send fails, anymail_status should contain initial values"""
        self.set_mock_response(status_code=500)
        sent = self.message.send(fail_silently=True)
        self.assertEqual(sent, 0)
        self.assertIsNone(self.message.anymail_status.status)
        self.assertIsNone(self.message.anymail_status.message_id)
        self.assertEqual(self.message.anymail_status.recipients, {})
        self.assertIsNone(self.message.anymail_status.esp_response)

    def test_json_serialization_errors(self):
        """Try to provide more information about non-json-serializable data"""
        self.message.metadata = {"total": Decimal("19.99")}
        with self.assertRaises(AnymailSerializationError) as cm:
            self.message.send()
        err = cm.exception
        self.assertIsInstance(err, TypeError)  # compatibility with json.dumps
        # our added context:
        self.assertIn("Don't know how to send this data to SendGrid", str(err))
        # original message:
        self.assertRegex(str(err), r"Decimal.*is not JSON serializable")


@tag("sendgrid")
@ignore_warnings(category=AnymailNotSupportedWarning)
class SendGridBackendRecipientsRefusedTests(SendGridBackendMockAPITestCase):
    """
    Should raise AnymailRecipientsRefused when *all* recipients are rejected or invalid
    """

    # SendGrid doesn't check email bounce or complaint lists at time of send --
    # it always just queues the message. You'll need to listen for the "rejected"
    # and "failed" events to detect refused recipients.
    pass  # not applicable to this backend


@tag("sendgrid")
@ignore_warnings(category=AnymailNotSupportedWarning)
class SendGridBackendSessionSharingTestCase(
    SessionSharingTestCases, SendGridBackendMockAPITestCase
):
    """Requests session sharing tests"""

    pass  # tests are defined in SessionSharingTestCases


@tag("sendgrid")
@override_settings(EMAIL_BACKEND="anymail.backends.sendgrid.EmailBackend")
@ignore_warnings(category=AnymailNotSupportedWarning)
class SendGridBackendImproperlyConfiguredTests(AnymailTestMixin, SimpleTestCase):
    """Test ESP backend without required settings in place"""

    def test_missing_auth(self):
        with self.assertRaisesRegex(AnymailConfigurationError, r"\bSENDGRID_API_KEY\b"):
            mail.send_mail("Subject", "Message", "from@example.com", ["to@example.com"])


@tag("sendgrid")
@override_settings(EMAIL_BACKEND="anymail.backends.sendgrid.EmailBackend")
@ignore_warnings(category=AnymailNotSupportedWarning)
class SendGridBackendDisallowsV2Tests(AnymailTestMixin, SimpleTestCase):
    """Using v2-API-only features should cause errors with v3 backend"""

    @override_settings(
        ANYMAIL={"SENDGRID_USERNAME": "sg_username", "SENDGRID_PASSWORD": "sg_password"}
    )
    def test_user_pass_auth(self):
        """Make sure v2-only USERNAME/PASSWORD auth raises error"""
        with self.assertRaisesMessage(
            AnymailConfigurationError,
            "SendGrid v3 API doesn't support username/password auth;"
            " Please change to API key.",
        ):
            mail.send_mail("Subject", "Message", "from@example.com", ["to@example.com"])

    @override_settings(ANYMAIL={"SENDGRID_API_KEY": "test_api_key"})
    def test_esp_extra_smtpapi(self):
        """x-smtpapi in the esp_extra indicates a desire to use the v2 api"""
        message = mail.EmailMessage(
            "Subject", "Body", "from@example.com", ["to@example.com"]
        )
        message.esp_extra = {"x-smtpapi": {"asm_group_id": 1}}
        with self.assertRaisesMessage(
            AnymailConfigurationError,
            "You are attempting to use SendGrid v2 API-style x-smtpapi params with the"
            " SendGrid v3 API. Please update your `esp_extra` to the new API.",
        ):
            message.send()



import base64
import mimetypes
import re
from base64 import b64encode
from collections.abc import Mapping, MutableMapping
from copy import copy, deepcopy
from email.mime.base import MIMEBase
from email.utils import formatdate, getaddresses, parsedate_to_datetime, unquote
from urllib.parse import urlsplit, urlunsplit

from django.conf import settings
from django.core.mail.message import DEFAULT_ATTACHMENT_MIME_TYPE, sanitize_address
from django.utils.encoding import force_str
from django.utils.functional import Promise
from requests.structures import CaseInsensitiveDict

from .exceptions import AnymailConfigurationError, AnymailInvalidAddress

BASIC_NUMERIC_TYPES = (int, float)


UNSET = type("UNSET", (object,), {})  # Used as non-None default value


def concat_lists(*args):
    """
    Combines all non-UNSET args, by concatenating lists (or sequence-like types).
    Does not modify any args.

    >>> concat_lists([1, 2], UNSET, [3, 4], UNSET)
    [1, 2, 3, 4]
    >>> concat_lists([1, 2], None, [3, 4])  # None suppresses earlier args
    [3, 4]
    >>> concat_lists()
    UNSET

    """
    result = UNSET
    for value in args:
        if value is None:
            # None is a request to suppress any earlier values
            result = UNSET
        elif value is not UNSET:
            if result is UNSET:
                result = list(value)
            else:
                result = result + list(value)  # concatenate sequence-like
    return result


def merge_dicts_shallow(*args):
    """
    Shallow-merges all non-UNSET args.
    Does not modify any args.

    >>> merge_dicts_shallow({'a': 1, 'b': 2}, UNSET, {'b': 3, 'c': 4}, UNSET)
    {'a': 1, 'b': 3, 'c': 4}
    >>> merge_dicts_shallow({'a': {'a1': 1, 'a2': 2}}, {'a': {'a1': 11, 'a3': 33}})
    {'a': {'a1': 11, 'a3': 33}}
    >>> merge_dicts_shallow({'a': 1}, None, {'b': 2})  # None suppresses earlier args
    {'b': 2}
    >>> merge_dicts_shallow()
    UNSET

    """
    result = UNSET
    for value in args:
        if value is None:
            # None is a request to suppress any earlier values
            result = UNSET
        elif value is not UNSET:
            if result is UNSET:
                result = copy(value)
            else:
                result.update(value)
    return result


def merge_dicts_deep(*args):
    """
    Deep-merges all non-UNSET args.
    Does not modify any args.

    >>> merge_dicts_deep({'a': 1, 'b': 2}, UNSET, {'b': 3, 'c': 4}, UNSET)
    {'a': 1, 'b': 3, 'c': 4}
    >>> merge_dicts_deep({'a': {'a1': 1, 'a2': 2}}, {'a': {'a1': 11, 'a3': 33}})
    {'a': {'a1': 11, 'a2': 2, 'a3': 33}}
    >>> merge_dicts_deep({'a': 1}, None, {'b': 2})  # None suppresses earlier args
    {'b': 2}
    >>> merge_dicts_deep()
    UNSET

    """
    result = UNSET
    for value in args:
        if value is None:
            # None is a request to suppress any earlier values
            result = UNSET
        elif value is not UNSET:
            if result is UNSET:
                result = deepcopy(value)
            else:
                update_deep(result, value)
    return result


def merge_dicts_one_level(*args):
    """
    Mixture of merge_dicts_deep and merge_dicts_shallow:
    Deep merges first level, shallow merges second level.
    Does not modify any args.

    (Useful for {"email": {options...}, ...} style dicts,
    like merge_data: shallow merges the options for each email.)
    """
    result = UNSET
    for value in args:
        if value is None:
            # None is a request to suppress any earlier values
            result = UNSET
        elif value is not UNSET:
            if result is UNSET:
                result = {}
            for k, v in value.items():
                result.setdefault(k, {}).update(v)
    return result


def last(*args):
    """Returns the last of its args which is not UNSET.

    >>> last(1, 2, UNSET, 3, UNSET, UNSET)
    3
    >>> last(1, 2, None, UNSET)  # None suppresses earlier args
    UNSET
    >>> last()
    UNSET

    """
    for value in reversed(args):
        if value is None:
            # None is a request to suppress any earlier values
            return UNSET
        elif value is not UNSET:
            return value
    return UNSET


def getfirst(dct, keys, default=UNSET):
    """Returns the value of the first of keys found in dict dct.

    >>> getfirst({'a': 1, 'b': 2}, ['c', 'a'])
    1
    >>> getfirst({'a': 1, 'b': 2}, ['b', 'a'])
    2
    >>> getfirst({'a': 1, 'b': 2}, ['c'])
    KeyError
    >>> getfirst({'a': 1, 'b': 2}, ['c'], None)
    None
    """
    for key in keys:
        try:
            return dct[key]
        except KeyError:
            pass
    if default is UNSET:
        raise KeyError("None of %s found in dict" % ", ".join(keys))
    else:
        return default


def update_deep(dct, other):
    """Merge (recursively) keys and values from dict other into dict dct

    Works with dict-like objects: dct (and descendants) can be any MutableMapping,
    and other can be any Mapping
    """
    for key, value in other.items():
        if (
            key in dct
            and isinstance(dct[key], MutableMapping)
            and isinstance(value, Mapping)
        ):
            update_deep(dct[key], value)
        else:
            dct[key] = value
    # (like dict.update(), no return value)


def parse_address_list(address_list, field=None):
    """Returns a list of EmailAddress objects from strings in address_list.

    Essentially wraps :func:`email.utils.getaddresses` with better error
    messaging and more-useful output objects

    Note that the returned list might be longer than the address_list param,
    if any individual string contains multiple comma-separated addresses.

    :param list[str]|str|None|list[None] address_list:
        the address or addresses to parse
    :param str|None field:
        optional description of the source of these addresses, for error message
    :return list[:class:`EmailAddress`]:
    :raises :exc:`AnymailInvalidAddress`:
    """
    if isinstance(address_list, str) or is_lazy(address_list):
        address_list = [address_list]

    if address_list is None or address_list == [None]:
        return []

    # For consistency with Django's SMTP backend behavior, extract all addresses
    # from the list -- which may split comma-seperated strings into multiple addresses.
    # (See django.core.mail.message: EmailMessage.message to/cc/bcc/reply_to handling;
    # also logic for ADDRESS_HEADERS in forbid_multi_line_headers.)

    # resolve lazy strings:
    address_list_strings = [force_str(address) for address in address_list]
    name_email_pairs = getaddresses(address_list_strings)
    if name_email_pairs == [] and address_list_strings == [""]:
        name_email_pairs = [("", "")]  # getaddresses ignores a single empty string
    parsed = [
        EmailAddress(display_name=name, addr_spec=email)
        for (name, email) in name_email_pairs
    ]

    # Sanity-check, and raise useful errors
    for address in parsed:
        if address.username == "" or address.domain == "":
            # Django SMTP allows username-only emails,
            # but they're not meaningful with an ESP
            errmsg = (
                "Invalid email address '{problem}'" " parsed from '{source}'{where}."
            ).format(
                problem=address.addr_spec,
                source=", ".join(address_list_strings),
                where=" in `%s`" % field if field else "",
            )
            if len(parsed) > len(address_list):
                errmsg += " (Maybe missing quotes around a display-name?)"
            raise AnymailInvalidAddress(errmsg)

    return parsed


def parse_single_address(address, field=None):
    """Parses a single EmailAddress from str address, or raises AnymailInvalidAddress

    :param str address: the fully-formatted email str to parse
    :param str|None field:
        optional description of the source of this address, for error message
    :return :class:`EmailAddress`: if address contains a single email
    :raises :exc:`AnymailInvalidAddress`: if address contains no or multiple emails
    """
    parsed = parse_address_list([address], field=field)
    count = len(parsed)
    if count > 1:
        raise AnymailInvalidAddress(
            "Only one email address is allowed;"
            " found {count} in '{address}'{where}.".format(
                count=count, address=address, where=" in `%s`" % field if field else ""
            )
        )
    else:
        return parsed[0]


class EmailAddress:
    """A sanitized, complete email address with easy access
    to display-name, addr-spec (email), etc.

    Similar to Python 3.6+ email.headerregistry.Address

    Instance properties, all read-only:
    :ivar str display_name:
        the address's display-name portion (unqouted, unescaped),
        e.g., 'Display Name, Inc.'
    :ivar str addr_spec:
        the address's addr-spec portion (unquoted, unescaped),
        e.g., 'user@example.com'
    :ivar str username:
        the local part (before the '@') of the addr-spec,
        e.g., 'user'
    :ivar str domain:
        the domain part (after the '@') of the addr-spec,
        e.g., 'example.com'

    :ivar str address:
        the fully-formatted address, with any necessary quoting and escaping,
        e.g., '"Display Name, Inc." <user@example.com>'
        (also available as `str(EmailAddress)`)
    """

    def __init__(self, display_name="", addr_spec=None):
        self._address = None  # lazy formatted address
        if addr_spec is None:
            try:
                display_name, addr_spec = display_name  # unpack (name,addr) tuple
            except ValueError:
                pass

        # ESPs should clean or reject addresses containing newlines, but some
        # extra protection can't hurt (and it seems to be a common oversight)
        if "\n" in display_name or "\r" in display_name:
            raise ValueError("EmailAddress display_name cannot contain newlines")
        if "\n" in addr_spec or "\r" in addr_spec:
            raise ValueError("EmailAddress addr_spec cannot contain newlines")

        self.display_name = display_name
        self.addr_spec = addr_spec
        try:
            self.username, self.domain = addr_spec.split("@", 1)
            # do we need to unquote username?
        except ValueError:
            self.username = addr_spec
            self.domain = ""

    def __repr__(self):
        return "EmailAddress({display_name!r}, {addr_spec!r})".format(
            display_name=self.display_name, addr_spec=self.addr_spec
        )

    @property
    def address(self):
        if self._address is None:
            # (you might be tempted to use `encoding=settings.DEFAULT_CHARSET` here,
            # but that always forces the display-name to quoted-printable/base64,
            # even when simple ascii would work fine--and be more readable)
            self._address = self.formataddr()
        return self._address

    def formataddr(self, encoding=None):
        """Return a fully-formatted email address, using encoding.

        This is essentially the same as :func:`email.utils.formataddr`
        on the EmailAddress's name and email properties, but uses
        Django's :func:`~django.core.mail.message.sanitize_address`
        for consistent handling of encoding (a.k.a. charset) and
        proper handling of IDN domain portions.

        :param str|None encoding:
            the charset to use for the display-name portion;
            default None uses ascii if possible, else 'utf-8'
            (quoted-printable utf-8/base64)
        """
        sanitized = sanitize_address((self.display_name, self.addr_spec), encoding)
        # sanitize_address() can introduce FWS with a long, non-ASCII display name.
        # Must unfold it:
        return re.sub(r"(\r|\n|\r\n)[ \t]", "", sanitized)

    def __str__(self):
        return self.address


class Attachment:
    """A normalized EmailMessage.attachments item with additional functionality

    Normalized to have these properties:
    name: attachment filename; may be None
    content: bytestream
    mimetype: the content type; guessed if not explicit
    inline: bool, True if attachment has a Content-ID header
    content_id: for inline, the Content-ID (*with* <>); may be None
    cid: for inline, the Content-ID *without* <>; may be empty string
    """

    def __init__(self, attachment, encoding):
        # Note that an attachment can be either a tuple of (filename, content, mimetype)
        # or a MIMEBase object. (Also, both filename and mimetype may be missing.)
        self._attachment = attachment
        self.encoding = encoding  # or check attachment["Content-Encoding"] ???
        self.inline = False
        self.content_id = None
        self.cid = ""

        if isinstance(attachment, MIMEBase):
            self.name = attachment.get_filename()
            self.content = attachment.get_payload(decode=True)
            if self.content is None:
                self.content = attachment.as_bytes()
            self.mimetype = attachment.get_content_type()
            # Content-Type includes charset if provided
            self.content_type = attachment["Content-Type"]

            content_disposition = attachment.get_content_disposition()
            if content_disposition == "inline" or (
                not content_disposition and "Content-ID" in attachment
            ):
                self.inline = True
                self.content_id = attachment["Content-ID"]  # probably including <...>
                if self.content_id is not None:
                    self.cid = unquote(self.content_id)  # without the <, >
        else:
            (self.name, self.content, self.mimetype) = attachment
            self.content_type = self.mimetype

        self.name = force_non_lazy(self.name)
        self.content = force_non_lazy(self.content)

        # Guess missing mimetype from filename, borrowed from
        # django.core.mail.EmailMessage._create_attachment()
        if self.mimetype is None and self.name is not None:
            self.mimetype, _ = mimetypes.guess_type(self.name)
        if self.mimetype is None:
            self.mimetype = DEFAULT_ATTACHMENT_MIME_TYPE
        if self.content_type is None:
            self.content_type = self.mimetype

    def __repr__(self):
        details = [
            self.mimetype,
            "len={length}".format(length=len(self.content)),
        ]
        if self.name:
            details.append("name={name!r}".format(name=self.name))
        if self.inline:
            details.insert(0, "inline")
            details.append(
                "content_id={content_id!r}".format(content_id=self.content_id)
            )
        return "Attachment<{details}>".format(details=", ".join(details))

    @property
    def b64content(self):
        """Content encoded as a base64 ascii string"""
        content = self.content
        if isinstance(content, str):
            content = content.encode(self.encoding)
        return b64encode(content).decode("ascii")


def get_anymail_setting(
    name, default=UNSET, esp_name=None, kwargs=None, allow_bare=False
):
    """Returns an Anymail option from kwargs or Django settings.

    Returns first of:
    - kwargs[name] -- e.g., kwargs['api_key'] -- and name key will be popped from kwargs
    - settings.ANYMAIL['<ESP_NAME>_<NAME>'] -- e.g., settings.ANYMAIL['MAILGUN_API_KEY']
    - settings.ANYMAIL_<ESP_NAME>_<NAME> -- e.g., settings.ANYMAIL_MAILGUN_API_KEY
    - settings.<ESP_NAME>_<NAME> (only if allow_bare) -- e.g., settings.MAILGUN_API_KEY
    - default if provided; else raises AnymailConfigurationError

    If allow_bare, allows settings.<ESP_NAME>_<NAME> without the ANYMAIL_ prefix:
    ANYMAIL = { "MAILGUN_API_KEY": "xyz", ... }
    ANYMAIL_MAILGUN_API_KEY = "xyz"
    MAILGUN_API_KEY = "xyz"
    """

    try:
        value = kwargs.pop(name)
        if name in ["username", "password"]:
            # Work around a problem in django.core.mail.send_mail, which calls
            # get_connection(... username=None, password=None) by default.
            # We need to ignore those None defaults (else settings like
            # 'SENDGRID_USERNAME' get unintentionally overridden from kwargs).
            if value is not None:
                return value
        else:
            return value
    except (AttributeError, KeyError):
        pass

    if esp_name is not None:
        setting = "{}_{}".format(esp_name.upper().replace(" ", "_"), name.upper())
    else:
        setting = name.upper()
    anymail_setting = "ANYMAIL_%s" % setting

    try:
        return settings.ANYMAIL[setting]
    except (AttributeError, KeyError):
        try:
            return getattr(settings, anymail_setting)
        except AttributeError:
            if allow_bare:
                try:
                    return getattr(settings, setting)
                except AttributeError:
                    pass
            if default is UNSET:
                message = "You must set %s or ANYMAIL = {'%s': ...}" % (
                    anymail_setting,
                    setting,
                )
                if allow_bare:
                    message += " or %s" % setting
                message += " in your Django settings"
                raise AnymailConfigurationError(message) from None
            else:
                return default


def collect_all_methods(cls, method_name):
    """Return list of all `method_name` methods for cls and its superclass chain.

    List is in MRO order, with no duplicates. Methods are unbound.

    (This is used to simplify mixins and subclasses that contribute to a method set,
    without requiring superclass chaining, and without requiring cooperating
    superclasses.)
    """
    methods = []
    for ancestor in cls.__mro__:
        try:
            validator = getattr(ancestor, method_name)
        except AttributeError:
            pass
        else:
            if validator not in methods:
                methods.append(validator)
    return methods


def querydict_getfirst(qdict, field, default=UNSET):
    """
    Like :func:`django.http.QueryDict.get`,
    but returns *first* value of multi-valued field.

    >>> from django.http import QueryDict
    >>> q = QueryDict('a=1&a=2&a=3')
    >>> querydict_getfirst(q, 'a')
    '1'
    >>> q.get('a')
    '3'
    >>> q['a']
    '3'

    You can bind this to a QueryDict instance using the "descriptor protocol":
    >>> q.getfirst = querydict_getfirst.__get__(q)
    >>> q.getfirst('a')
    '1'
    """
    # (Why not instead define a QueryDict subclass with this method? Because there's
    # no simple way to efficiently initialize a QueryDict subclass with the contents
    # of an existing instance.)
    values = qdict.getlist(field)
    if len(values) > 0:
        return values[0]
    elif default is not UNSET:
        return default
    else:
        return qdict[field]  # raise appropriate KeyError


def rfc2822date(dt):
    """Turn a datetime into a date string as specified in RFC 2822."""
    # This is almost the equivalent of Python's email.utils.format_datetime,
    # but treats naive datetimes as local rather than "UTC with no information ..."
    timeval = dt.timestamp()
    return formatdate(timeval, usegmt=True)


def angle_wrap(s):
    """Return s surrounded by angle brackets, added only if necessary"""
    # This is the inverse behavior of email.utils.unquote
    # (which you might think email.utils.quote would do, but it doesn't)
    if len(s) > 0:
        if s[0] != "<":
            s = "<" + s
        if s[-1] != ">":
            s = s + ">"
    return s


def is_lazy(obj):
    """Return True if obj is a Django lazy object."""
    # See django.utils.functional.lazy. (This appears to be preferred
    # to checking for `not isinstance(obj, str)`.)
    return isinstance(obj, Promise)


def force_non_lazy(obj):
    """
    If obj is a Django lazy object, return it coerced to text;
    otherwise return it unchanged.

    (Similar to django.utils.encoding.force_text, but doesn't alter non-text objects.)
    """
    if is_lazy(obj):
        return str(obj)

    return obj


def force_non_lazy_list(obj):
    """Return a (shallow) copy of sequence obj, with all values forced non-lazy."""
    try:
        return [force_non_lazy(item) for item in obj]
    except (AttributeError, TypeError):
        return force_non_lazy(obj)


def force_non_lazy_dict(obj):
    """Return a (deep) copy of dict obj, with all values forced non-lazy."""
    try:
        return {key: force_non_lazy_dict(value) for key, value in obj.items()}
    except (AttributeError, TypeError):
        return force_non_lazy(obj)


def get_request_basic_auth(request):
    """Returns HTTP basic auth string sent with request, or None.

    If request includes basic auth, result is string 'username:password'.
    """
    try:
        authtype, authdata = request.META["HTTP_AUTHORIZATION"].split()
        if authtype.lower() == "basic":
            return base64.b64decode(authdata).decode("utf-8")
    except (IndexError, KeyError, TypeError, ValueError):
        pass
    return None


def get_request_uri(request):
    """Returns the "exact" url used to call request.

    Like :func:`django.http.request.HTTPRequest.build_absolute_uri`,
    but also inlines HTTP basic auth, if present.
    """
    url = request.build_absolute_uri()
    basic_auth = get_request_basic_auth(request)
    if basic_auth is not None:
        # must reassemble url with auth
        parts = urlsplit(url)
        url = urlunsplit(
            (
                parts.scheme,
                basic_auth + "@" + parts.netloc,
                parts.path,
                parts.query,
                parts.fragment,
            )
        )
    return url


def parse_rfc2822date(s):
    """Parses an RFC-2822 formatted date string into a datetime.datetime

    Returns None if string isn't parseable. Returned datetime will be naive
    if string doesn't include known timezone offset; aware if it does.

    (Same as Python 3 email.utils.parsedate_to_datetime, with improved
    handling for unparseable date strings.)
    """
    try:
        return parsedate_to_datetime(s)
    except (IndexError, TypeError, ValueError):
        # despite the docs, parsedate_to_datetime often dies on unparseable input
        return None


class CaseInsensitiveCasePreservingDict(CaseInsensitiveDict):
    """A dict with case-insensitive keys, which preserves the *first* key set.

    >>> cicpd = CaseInsensitiveCasePreservingDict()
    >>> cicpd["Accept"] = "application/text+xml"
    >>> cicpd["accEPT"] = "application/json"
    >>> cicpd["accept"]
    "application/json"
    >>> cicpd.keys()
    ["Accept"]

    Compare to CaseInsensitiveDict, which preserves *last* key set:
    >>> cid = CaseInsensitiveCasePreservingDict()
    >>> cid["Accept"] = "application/text+xml"
    >>> cid["accEPT"] = "application/json"
    >>> cid.keys()
    ["accEPT"]
    """

    def __setitem__(self, key, value):
        _k = key.lower()
        try:
            # retrieve earlier matching key, if any
            key, _ = self._store[_k]
        except KeyError:
            pass
        self._store[_k] = (key, value)

    def copy(self):
        return self.__class__(self._store.values())



import json
from datetime import datetime, timezone
from unittest.mock import ANY

from django.test import tag

from anymail.signals import AnymailTrackingEvent
from anymail.webhooks.mailjet import MailjetTrackingWebhookView

from .webhook_cases import WebhookBasicAuthTestCase, WebhookTestCase


@tag("mailjet")
class MailjetWebhookSecurityTestCase(WebhookBasicAuthTestCase):
    def call_webhook(self):
        return self.client.post(
            "/anymail/mailjet/tracking/",
            content_type="application/json",
            data=json.dumps([]),
        )

    # Actual tests are in WebhookBasicAuthTestCase


@tag("mailjet")
class MailjetDeliveryTestCase(WebhookTestCase):
    def test_sent_event(self):
        # Mailjet's "sent" event indicates receiving MTA has accepted message;
        # Anymail calls this "delivered"
        raw_events = [
            {
                "event": "sent",
                "time": 1498093527,
                "MessageID": 12345678901234567,
                "email": "recipient@example.com",
                "mj_campaign_id": 1234567890,
                "mj_contact_id": 9876543210,
                "customcampaign": "tag1",
                "mj_message_id": "12345678901234567",
                "smtp_reply": "sent (250 2.0.0 OK 1498093527 a67bc12345def.22 - gsmtp)",
                "Payload": '{"meta1": "simple string", "meta2": 2}',
            }
        ]
        response = self.client.post(
            "/anymail/mailjet/tracking/",
            content_type="application/json",
            data=json.dumps(raw_events),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MailjetTrackingWebhookView,
            event=ANY,
            esp_name="Mailjet",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "delivered")
        self.assertEqual(
            event.timestamp, datetime(2017, 6, 22, 1, 5, 27, tzinfo=timezone.utc)
        )
        self.assertEqual(event.esp_event, raw_events[0])
        self.assertEqual(
            event.mta_response,
            "sent (250 2.0.0 OK 1498093527 a67bc12345def.22 - gsmtp)",
        )
        # message_id converted to str (matching backend status):
        self.assertEqual(event.message_id, "12345678901234567")
        self.assertEqual(event.recipient, "recipient@example.com")
        self.assertEqual(event.tags, ["tag1"])
        self.assertEqual(event.metadata, {"meta1": "simple string", "meta2": 2})

    def test_open_event(self):
        raw_events = [
            {
                "event": "open",
                "time": 1498093527,
                "MessageID": 12345678901234567,
                "email": "recipient@example.com",
                "mj_campaign_id": 1234567890,
                "mj_contact_id": 9876543210,
                "customcampaign": "",
                "ip": "192.168.100.100",
                "geo": "US",
                "agent": "Mozilla/5.0 (via ggpht.com GoogleImageProxy)",
                "Payload": "",
            }
        ]
        response = self.client.post(
            "/anymail/mailjet/tracking/",
            content_type="application/json",
            data=json.dumps(raw_events),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MailjetTrackingWebhookView,
            event=ANY,
            esp_name="Mailjet",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "opened")
        self.assertEqual(event.message_id, "12345678901234567")
        self.assertEqual(event.recipient, "recipient@example.com")
        self.assertEqual(
            event.user_agent, "Mozilla/5.0 (via ggpht.com GoogleImageProxy)"
        )
        self.assertEqual(event.tags, [])
        self.assertEqual(event.metadata, {})

    def test_click_event(self):
        raw_events = [
            {
                "event": "open",
                "time": 1498093527,
                "MessageID": 12345678901234567,
                "email": "recipient@example.com",
                "mj_campaign_id": 1234567890,
                "mj_contact_id": 9876543210,
                "customcampaign": "",
                "url": "http://example.com",
                "ip": "192.168.100.100",
                "geo": "US",
                "agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_5)"
                " Chrome/58.0.3029.110",
            }
        ]
        response = self.client.post(
            "/anymail/mailjet/tracking/",
            content_type="application/json",
            data=json.dumps(raw_events),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MailjetTrackingWebhookView,
            event=ANY,
            esp_name="Mailjet",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "opened")
        self.assertEqual(event.message_id, "12345678901234567")
        self.assertEqual(event.recipient, "recipient@example.com")
        self.assertEqual(
            event.user_agent,
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_5) Chrome/58.0.3029.110",
        )
        self.assertEqual(event.click_url, "http://example.com")
        self.assertEqual(event.tags, [])
        self.assertEqual(event.metadata, {})

    def test_bounce_event(self):
        raw_events = [
            {
                "event": "bounce",
                "time": 1498093527,
                "MessageID": 12345678901234567,
                "email": "invalid@invalid",
                "mj_campaign_id": 1234567890,
                "mj_contact_id": 9876543210,
                "customcampaign": "",
                "blocked": True,
                "hard_bounce": True,
                "error_related_to": "domain",
                "error": "invalid domain",
            }
        ]
        response = self.client.post(
            "/anymail/mailjet/tracking/",
            content_type="application/json",
            data=json.dumps(raw_events),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MailjetTrackingWebhookView,
            event=ANY,
            esp_name="Mailjet",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "bounced")
        self.assertEqual(event.message_id, "12345678901234567")
        self.assertEqual(event.recipient, "invalid@invalid")
        self.assertEqual(event.reject_reason, "bounced")
        self.assertEqual(event.mta_response, None)

    def test_blocked_event(self):
        raw_events = [
            {
                "event": "blocked",
                "time": 1498093527,
                "MessageID": 12345678901234567,
                "email": "bad@example.com",
                "mj_campaign_id": 0,
                "mj_contact_id": 9876543210,
                "customcampaign": "",
                "error_related_to": "domain",
                "error": "typofix",
            }
        ]
        response = self.client.post(
            "/anymail/mailjet/tracking/",
            content_type="application/json",
            data=json.dumps(raw_events),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MailjetTrackingWebhookView,
            event=ANY,
            esp_name="Mailjet",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "rejected")
        self.assertEqual(event.message_id, "12345678901234567")
        self.assertEqual(event.recipient, "bad@example.com")
        self.assertEqual(event.reject_reason, "invalid")
        self.assertEqual(event.mta_response, None)

    def test_spam_event(self):
        raw_events = [
            {
                "event": "spam",
                "time": 1498093527,
                "MessageID": 12345678901234567,
                "email": "spam@example.com",
                "mj_campaign_id": 1234567890,
                "mj_contact_id": 9876543210,
                "customcampaign": "",
                "source": "greylisted",
            }
        ]
        response = self.client.post(
            "/anymail/mailjet/tracking/",
            content_type="application/json",
            data=json.dumps(raw_events),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MailjetTrackingWebhookView,
            event=ANY,
            esp_name="Mailjet",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "complained")
        self.assertEqual(event.message_id, "12345678901234567")
        self.assertEqual(event.recipient, "spam@example.com")

    def test_unsub_event(self):
        raw_events = [
            {
                "event": "unsub",
                "time": 1498093527,
                "MessageID": 12345678901234567,
                "email": "recipient@example.com",
                "mj_campaign_id": 1234567890,
                "mj_contact_id": 9876543210,
                "customcampaign": "",
                "mj_list_id": 0,
                "ip": "127.0.0.4",
                "geo": "",
                "agent": "List-Unsubscribe",
            }
        ]
        response = self.client.post(
            "/anymail/mailjet/tracking/",
            content_type="application/json",
            data=json.dumps(raw_events),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MailjetTrackingWebhookView,
            event=ANY,
            esp_name="Mailjet",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "unsubscribed")
        self.assertEqual(event.message_id, "12345678901234567")
        self.assertEqual(event.recipient, "recipient@example.com")

    def test_bounced_greylist_event(self):
        # greylist "bounce" should be reported as "deferred" (will be retried later)
        raw_events = [
            {
                "event": "bounce",
                "time": 1498093527,
                "MessageID": 12345678901234567,
                "email": "protected@example.com",
                "mj_campaign_id": 1234567890,
                "mj_contact_id": 9876543210,
                "customcampaign": "",
                "blocked": True,
                "hard_bounce": False,
                "error_related_to": "domain",
                "error": "greylisted",
            }
        ]
        response = self.client.post(
            "/anymail/mailjet/tracking/",
            content_type="application/json",
            data=json.dumps(raw_events),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MailjetTrackingWebhookView,
            event=ANY,
            esp_name="Mailjet",
        )
        event = kwargs["event"]
        self.assertEqual(event.event_type, "deferred")
        self.assertEqual(event.message_id, "12345678901234567")
        self.assertEqual(event.recipient, "protected@example.com")
        self.assertEqual(event.reject_reason, "other")

    def test_non_grouped_event(self):
        # If you don't enable "group events" on a webhook, Mailjet sends a single bare
        # event (not a list of one event, despite what the docs say).
        raw_event = {
            "event": "sent",
            "time": 1498093527,
            "MessageID": 12345678901234567,
            "email": "recipient@example.com",
            "mj_campaign_id": 1234567890,
            "mj_contact_id": 9876543210,
            "customcampaign": "",
            "mj_message_id": "12345678901234567",
            "smtp_reply": "sent (250 2.0.0 OK 1498093527 a67bc12345def.22 - gsmtp)",
            "Payload": "",
        }
        response = self.client.post(
            "/anymail/mailjet/tracking/",
            content_type="application/json",
            data=json.dumps(raw_event),
        )
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.tracking_handler,
            sender=MailjetTrackingWebhookView,
            event=ANY,
            esp_name="Mailjet",
        )
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailTrackingEvent)
        self.assertEqual(event.event_type, "delivered")
        self.assertEqual(
            event.timestamp, datetime(2017, 6, 22, 1, 5, 27, tzinfo=timezone.utc)
        )
        self.assertEqual(event.esp_event, raw_event)
        self.assertEqual(
            event.mta_response,
            "sent (250 2.0.0 OK 1498093527 a67bc12345def.22 - gsmtp)",
        )
        # message_id converted to str (matching backend status)
        self.assertEqual(event.message_id, "12345678901234567")
        self.assertEqual(event.recipient, "recipient@example.com")



import json
from datetime import datetime, timezone
from textwrap import dedent
from unittest.mock import ANY

from django.core.exceptions import ImproperlyConfigured
from django.test import override_settings, tag

from anymail.exceptions import AnymailConfigurationError
from anymail.inbound import AnymailInboundMessage
from anymail.signals import AnymailInboundEvent
from anymail.webhooks.mailersend import MailerSendInboundWebhookView

from .test_mailersend_webhooks import (
    TEST_WEBHOOK_SIGNING_SECRET,
    MailerSendWebhookTestCase,
)
from .utils import sample_image_content
from .webhook_cases import WebhookBasicAuthTestCase


@tag("mailersend")
@override_settings(ANYMAIL_MAILERSEND_INBOUND_SECRET=TEST_WEBHOOK_SIGNING_SECRET)
class MailerSendInboundSecurityTestCase(
    MailerSendWebhookTestCase, WebhookBasicAuthTestCase
):
    should_warn_if_no_auth = False  # because we check webhook signature

    def call_webhook(self):
        return self.client_post_signed(
            "/anymail/mailersend/inbound/",
            {"type": "inbound.message", "data": {"raw": "..."}},
            secret=TEST_WEBHOOK_SIGNING_SECRET,
        )

    # Additional tests are in WebhookBasicAuthTestCase

    def test_verifies_correct_signature(self):
        response = self.client_post_signed(
            "/anymail/mailersend/inbound/",
            {"type": "inbound.message", "data": {"raw": "..."}},
            secret=TEST_WEBHOOK_SIGNING_SECRET,
        )
        self.assertEqual(response.status_code, 200)

    def test_verifies_missing_signature(self):
        response = self.client.post(
            "/anymail/mailersend/inbound/",
            content_type="application/json",
            data=json.dumps({"type": "inbound.message", "data": {"raw": "..."}}),
        )
        self.assertEqual(response.status_code, 400)

    def test_verifies_bad_signature(self):
        # This also verifies that the error log references the correct setting to check.
        with self.assertLogs() as logs:
            response = self.client_post_signed(
                "/anymail/mailersend/inbound/",
                {"type": "inbound.message", "data": {"raw": "..."}},
                secret="wrong signing key",
            )
        # SuspiciousOperation causes 400 response (even in test client):
        self.assertEqual(response.status_code, 400)
        self.assertIn("check Anymail MAILERSEND_INBOUND_SECRET", logs.output[0])


@tag("mailersend")
class MailerSendInboundSettingsTestCase(MailerSendWebhookTestCase):
    def test_requires_inbound_secret(self):
        with self.assertRaisesMessage(
            ImproperlyConfigured, "MAILERSEND_INBOUND_SECRET"
        ):
            self.client_post_signed(
                "/anymail/mailersend/inbound/",
                {
                    "type": "inbound.message",
                    "data": {"object": "message", "raw": "..."},
                },
            )

    @override_settings(
        ANYMAIL={
            "MAILERSEND_INBOUND_SECRET": "inbound secret",
            "MAILERSEND_WEBHOOK_SIGNING_SECRET": "webhook secret",
        }
    )
    def test_webhook_signing_secret_is_different(self):
        response = self.client_post_signed(
            "/anymail/mailersend/inbound/",
            {
                "type": "inbound.message",
                "data": {"object": "message", "raw": "..."},
            },
            secret="inbound secret",
        )
        self.assertEqual(response.status_code, 200)

    @override_settings(ANYMAIL_MAILERSEND_INBOUND_SECRET="settings secret")
    def test_inbound_secret_view_params(self):
        """Webhook signing secret can be provided as a view param"""
        view = MailerSendInboundWebhookView.as_view(inbound_secret="view-level secret")
        view_instance = view.view_class(**view.view_initkwargs)
        self.assertEqual(view_instance.signing_secret, b"view-level secret")


@tag("mailersend")
@override_settings(ANYMAIL_MAILERSEND_INBOUND_SECRET=TEST_WEBHOOK_SIGNING_SECRET)
class MailerSendInboundTestCase(MailerSendWebhookTestCase):
    # Since Anymail just parses the raw MIME message through the Python email
    # package, there aren't really a lot of different cases to test here.
    # (We don't need to re-test the whole email.parser.)

    def test_inbound(self):
        # This is an actual (sanitized) inbound payload received from MailerSend:
        raw_event = {
            "type": "inbound.message",
            "inbound_id": "[inbound-route-id-redacted]",
            "url": "https://test.anymail.dev/anymail/mailersend/inbound/",
            "created_at": "2023-03-04T02:22:16.417935Z",
            "data": {
                "object": "message",
                "id": "6402ab57f79d39d7e10f2523",
                "recipients": {
                    "rcptTo": [{"email": "envelope-recipient@example.com"}],
                    "to": {
                        "raw": "Recipient <to@example.com>",
                        "data": [{"email": "to@example.com", "name": "Recipient"}],
                    },
                },
                "from": {
                    "email": "sender@example.org",
                    "name": "Sender Name",
                    "raw": "Sender Name <sender@example.org>",
                },
                "sender": {"email": "envelope-sender@example.org"},
                "subject": "Testing inbound \ud83c\udf0e",
                "date": "Fri, 3 Mar 2023 18:22:03 -0800",
                "headers": {
                    "X-Envelope-From": "<envelope-sender@example.org>",
                    # Multiple-instance headers appear as arrays:
                    "Received": [
                        "from example.org (mail.example.org [10.10.10.10])\r\n"
                        " by inbound.mailersend.net with ESMTPS id ...\r\n"
                        " Sat, 04 Mar 2023 02:22:15 +0000 (UTC)",
                        "by mail.example.org with SMTP id ...\r\n"
                        " for <envelope-recipient@example.com>;\r\n"
                        " Fri, 03 Mar 2023 18:22:15 -0800 (PST)",
                    ],
                    "DKIM-Signature": "v=1; a=rsa-sha256; c=relaxed/relaxed; ...",
                    "MIME-Version": "1.0",
                    "From": "Sender Name <sender@example.org>",
                    "Date": "Fri, 3 Mar 2023 18:22:03 -0800",
                    "Message-ID": "<AzjSdSHsmvXUeZGTPQ@mail.example.org>",
                    "Subject": "=?UTF-8?Q?Testing_inbound_=F0=9F=8C=8E?=",
                    "To": "Recipient <to@example.com>",
                    "Content-Type": 'multipart/mixed; boundary="000000000000e5575c05f609bab6"',
                },
                "text": "This is a *test*!\r\n\r\n[image: sample_image.png]\r\n",
                "html": (
                    "<p>This is a <b>test</b>!</p>"
                    '<img src="cid:ii_letc8ro50" alt="sample_image.png">'
                ),
                "raw": dedent(
                    """\
                    X-Envelope-From: <envelope-sender@example.org>
                    Received: from example.org (mail.example.org [10.10.10.10])
                     by inbound.mailersend.net with ESMTPS id ...
                     Sat, 04 Mar 2023 02:22:15 +0000 (UTC)
                    Received: by mail.example.org with SMTP id ...
                     for <envelope-recipient@example.com>;
                     Fri, 03 Mar 2023 18:22:15 -0800 (PST)
                    DKIM-Signature: v=1; a=rsa-sha256; c=relaxed/relaxed; ...
                    MIME-Version: 1.0
                    From: Sender Name <sender@example.org>
                    Date: Fri, 3 Mar 2023 18:22:03 -0800
                    Message-ID: <AzjSdSHsmvXUeZGTPQ@mail.example.org>
                    Subject: =?UTF-8?Q?Testing_inbound_=F0=9F=8C=8E?=
                    To: Recipient <to@example.com>
                    Content-Type: multipart/mixed; boundary="000000000000e5575c05f609bab6"

                    --000000000000e5575c05f609bab6
                    Content-Type: multipart/related; boundary="000000000000e5575b05f609bab5"

                    --000000000000e5575b05f609bab5
                    Content-Type: multipart/alternative; boundary="000000000000e5575a05f609bab4"

                    --000000000000e5575a05f609bab4
                    Content-Type: text/plain; charset="UTF-8"

                    This is a *test*!

                    [image: sample_image.png]

                    --000000000000e5575a05f609bab4
                    Content-Type: text/html; charset="UTF-8"

                    <p>This is a <b>test</b>!</p>
                    <img src="cid:ii_letc8ro50" alt="sample_image.png">

                    --000000000000e5575a05f609bab4--
                    --000000000000e5575b05f609bab5
                    Content-Type: image/png; name="sample_image.png"
                    Content-Disposition: inline; filename="sample_image.png"
                    Content-Transfer-Encoding: base64
                    Content-ID: <ii_letc8ro50>

                    iVBORw0KGgoAAAANSUhEUgAAACAAAAAgCAYAAABzenr0AAAABHNCSVQICAgIfAhkiAAAAAlwSFlz
                    AAALEgAACxIB0t1+/AAAABR0RVh0Q3JlYXRpb24gVGltZQAzLzEvMTNoZNRjAAAAHHRFWHRTb2Z0
                    d2FyZQBBZG9iZSBGaXJld29ya3MgQ1M1cbXjNgAAAZ1JREFUWIXtl7FKA0EQhr+TgIFgo5BXyBUp
                    fIGksLawUNAXWFFfwCJgBAtfIJFMLXgQn8BSwdpCiPcKAdOIoI2x2Dmyd7kYwXhp9odluX/uZv6d
                    nZu7DXowxiKZi0IAUHKCvxcsoAIEpST4IawVGb0Hb0BlpcigefACvAAvwAsoTTGGlwwzBAyivLUP
                    EZrOM10AhGOH2wWugVVlHoAdhJHrPC8DNR0JGsAAQ9mxNzBOMNjS4Qrq69U5EKmf12ywWVsQI4QI
                    IbCn3Gnmnk7uk1bokfooI7QRDlQIGCdzPwiYh0idtXNs2zq3UqwVEiDcu/R0DVjUnFpItuPSscfA
                    FXCGSfEAdZ2fVeQ68OjYWwi3ycVvMhABGwgfKXZScHeZ+4c6VzN8FbuYukvOykCs+z8PJ0xqIXYE
                    d4ALoKlVH2IIgUHWwd/6gNAFPjPcCPvKNTDcYAj1lXzKc7GIRrSZI6yJzcQ+dtV9bD+IkHThBj34
                    4j9/yYxupaQbXPJLNqsGFgeZ6qwpLP1b4AV4AV5AoKfjpR5OwR6VKwULCAC+AQV4W9Ps4uZQAAAA
                    AElFTkSuQmCC
                    --000000000000e5575b05f609bab5--
                    --000000000000e5575c05f609bab6
                    Content-Type: text/csv; charset="US-ASCII"; name="sample_data.csv"
                    Content-Disposition: attachment; filename="sample_data.csv"
                    Content-Transfer-Encoding: quoted-printable

                    Product,Price
                    Widget,33.20
                    --000000000000e5575c05f609bab6--"""
                ).replace("\n", "\r\n"),
                "attachments": [
                    {
                        "file_name": "sample_image.png",
                        "content_type": "image/png",
                        "content_disposition": "inline",
                        "content_id": "ii_letc8ro50",
                        "size": 579,
                        "content": (
                            "iVBORw0KGgoAAAANSUhEUgAAACAAAAAgCAYAAABzenr0AAAABHNCSVQICAgIfAhki"
                            "AAAAAlwSFlzAAALEgAACxIB0t1+/AAAABR0RVh0Q3JlYXRpb24gVGltZQAzLzEvMT"
                            "NoZNRjAAAAHHRFWHRTb2Z0d2FyZQBBZG9iZSBGaXJld29ya3MgQ1M1cbXjNgAAAZ1"
                            "JREFUWIXtl7FKA0EQhr+TgIFgo5BXyBUpfIGksLawUNAXWFFfwCJgBAtfIJFMLXgQ"
                            "n8BSwdpCiPcKAdOIoI2x2Dmyd7kYwXhp9odluX/uZv6dnZu7DXowxiKZi0IAUHKCv"
                            "xcsoAIEpST4IawVGb0Hb0BlpcigefACvAAvwAsoTTGGlwwzBAyivLUPEZrOM10AhG"
                            "OH2wWugVVlHoAdhJHrPC8DNR0JGsAAQ9mxNzBOMNjS4Qrq69U5EKmf12ywWVsQI4Q"
                            "IIbCn3Gnmnk7uk1bokfooI7QRDlQIGCdzPwiYh0idtXNs2zq3UqwVEiDcu/R0DVjU"
                            "nFpItuPSscfAFXCGSfEAdZ2fVeQ68OjYWwi3ycVvMhABGwgfKXZScHeZ+4c6VzN8F"
                            "buYukvOykCs+z8PJ0xqIXYEd4ALoKlVH2IIgUHWwd/6gNAFPjPcCPvKNTDcYAj1lX"
                            "zKc7GIRrSZI6yJzcQ+dtV9bD+IkHThBj344j9/yYxupaQbXPJLNqsGFgeZ6qwpLP1"
                            "b4AV4AV5AoKfjpR5OwR6VKwULCAC+AQV4W9Ps4uZQAAAAAElFTkSuQmCC"
                        ),
                    },
                    {
                        "file_name": "sample_data.csv",
                        "content_type": "text/csv",
                        "content_disposition": "attachment",
                        "size": 26,
                        "content": "UHJvZHVjdCxQcmljZQpXaWRnZXQsMzMuMjA=",
                    },
                ],
                "spf_check": {"code": "+", "value": None},
                "dkim_check": False,
                "created_at": "2023-03-04T02:22:15.525000Z",
            },
        }
        response = self.client_post_signed("/anymail/mailersend/inbound/", raw_event)
        self.assertEqual(response.status_code, 200)
        kwargs = self.assert_handler_called_once_with(
            self.inbound_handler,
            sender=MailerSendInboundWebhookView,
            event=ANY,
            esp_name="MailerSend",
        )
        # AnymailInboundEvent
        event = kwargs["event"]
        self.assertIsInstance(event, AnymailInboundEvent)
        self.assertEqual(event.event_type, "inbound")
        self.assertEqual(
            event.timestamp,
            # "2023-03-04T02:22:15.525000Z"
            datetime(2023, 3, 4, 2, 22, 15, microsecond=525000, tzinfo=timezone.utc),
        )
        self.assertEqual(event.event_id, "6402ab57f79d39d7e10f2523")
        self.assertIsInstance(event.message, AnymailInboundMessage)

        # (The raw_event subject contains a "\N{EARTH GLOBE AMERICAS}" (🌎)
        # character in the escaped form "\ud83c\udf0e", which won't compare equal
        # until unescaped. Passing through json dumps/loads resolves the escapes.)
        self.assertEqual(event.esp_event, json.loads(json.dumps(raw_event)))

        # AnymailInboundMessage - convenience properties
        message = event.message

        self.assertEqual(message.from_email.display_name, "Sender Name")
        self.assertEqual(message.from_email.addr_spec, "sender@example.org")
        self.assertEqual(str(message.to[0]), "Recipient <to@example.com>")
        self.assertEqual(message.subject, "Testing inbound 🌎")
        self.assertEqual(message.date.isoformat(" "), "2023-03-03 18:22:03-08:00")
        self.assertEqual(
            message.text, "This is a *test*!\r\n\r\n[image: sample_image.png]\r\n"
        )
        self.assertHTMLEqual(
            message.html,
            "<p>This is a <b>test</b>!</p>"
            '<img src="cid:ii_letc8ro50" alt="sample_image.png">',
        )

        self.assertEqual(message.envelope_sender, "envelope-sender@example.org")
        self.assertEqual(message.envelope_recipient, "envelope-recipient@example.com")

        # MailerSend inbound doesn't provide these:
        self.assertIsNone(message.stripped_text)
        self.assertIsNone(message.stripped_html)
        self.assertIsNone(message.spam_detected)
        self.assertIsNone(message.spam_score)

        # AnymailInboundMessage - other headers
        self.assertEqual(message["Message-ID"], "<AzjSdSHsmvXUeZGTPQ@mail.example.org>")
        self.assertEqual(
            message.get_all("Received"),
            [
                "from example.org (mail.example.org [10.10.10.10]) by inbound.mailersend.net"
                " with ESMTPS id ... Sat, 04 Mar 2023 02:22:15 +0000 (UTC)",
                "by mail.example.org with SMTP id ... for <envelope-recipient@example.com>;"
                " Fri, 03 Mar 2023 18:22:15 -0800 (PST)",
            ],
        )

        inlines = message.content_id_map
        self.assertEqual(len(inlines), 1)
        inline = inlines["ii_letc8ro50"]
        self.assertEqual(inline.get_filename(), "sample_image.png")
        self.assertEqual(inline.get_content_type(), "image/png")
        self.assertEqual(inline.get_content_bytes(), sample_image_content())

        attachments = message.attachments
        self.assertEqual(len(attachments), 1)
        self.assertEqual(attachments[0].get_filename(), "sample_data.csv")
        self.assertEqual(attachments[0].get_content_type(), "text/csv")
        self.assertEqual(
            attachments[0].get_content_text(), "Product,Price\r\nWidget,33.20"
        )

    def test_misconfigured_inbound(self):
        errmsg = (
            "You seem to have set MailerSend's *activity.sent* webhook"
            " to Anymail's MailerSend *inbound* webhook URL."
        )
        with self.assertRaisesMessage(AnymailConfigurationError, errmsg):
            self.client_post_signed(
                "/anymail/mailersend/inbound/",
                {
                    "type": "activity.sent",
                    "data": {"object": "activity", "type": "sent"},
                },
            )
